"""Adds support for "field" data types

While not "file-formats", in some workflows, particularly when interacting with a data
store such as in Arcana, it is useful to interchangeably interact with file and field
data, so the classes in this module are provided to support these use cases.
"""
import decimal
import typing as ty
import attrs
from fileformats.core import Field
from fileformats.core.mixin import WithClassifiers
from fileformats.core.exceptions import FormatMismatchError


class Singular(Field):
    pass


class LogicalMixin:
    def __bool__(self):
        return bool(self.value)

    def __and__(self, other):
        return self.value and other.value

    def __or__(self, other):
        return self.value or other.value

    def __not__(self):
        return not self.value

    def __invert__(self):
        return not self.value


class ScalarMixin(LogicalMixin):
    def __add__(self, other):
        return self.value + other

    def __sub__(self, other):
        return self.value - other

    def __mul__(self, other):
        return self.value * other

    def __truediv__(self, other):
        return self.value / other

    def __floordiv__(self, other):
        return self.value // other

    def __mod__(self, other):
        return self.value % other

    def __pow__(self, other):
        return self.value**other

    def __neg__(self):
        return -self.value

    def __pos__(self):
        return +self.value

    def __abs__(self):
        return abs(self.value)


def text_converter(value):
    try:
        return str(value)
    except ValueError as e:
        raise FormatMismatchError(str(e)) from None


def integer_converter(value):
    if isinstance(value, float):
        raise FormatMismatchError(
            f"Cannot convert float ({value}) to integer field without potential loss "
            "of information"
        )
    try:
        return int(value)
    except ValueError as e:
        raise FormatMismatchError(str(e)) from None


def decimal_converter(value):
    if isinstance(value, Decimal):
        return value.value
    try:
        return decimal.Decimal(value)
    except decimal.InvalidOperation as e:
        raise FormatMismatchError(str(e)) from None


def boolean_converter(value):
    if isinstance(value, str):
        if value.lower() in ("true", "1", "yes"):
            value = True
        elif value.lower() in ("false", "0", "no"):
            value = False
        else:
            raise FormatMismatchError(
                f"Cannot convert string '{value}' to boolean value"
            )
    else:
        try:
            value = bool(value)
        except ValueError as e:
            raise FormatMismatchError(str(e)) from None
    return value


def array_converter(value):
    if isinstance(value, str):
        if value.startswith("[") and value.endswith("]"):
            value = value[1:-1]
        elif value.startswith("[") or value.endswith("]"):
            raise FormatMismatchError(f"Unmatched brackets in array field {value}")
        value = tuple(v.strip() for v in value.split(","))
    else:
        try:
            value = tuple(value)
        except ValueError as e:
            raise FormatMismatchError(str(e)) from None
    return value


@attrs.define(repr=False)
class Text(Singular):

    value: str = attrs.field(converter=text_converter)

    primitive = str

    def __hash__(self):
        return hash(self.value)

    def __repr__(self):
        return f'{self._type_name}("{self.value}")'


@attrs.define(repr=False)
class Integer(Singular, ScalarMixin):

    value: int = attrs.field(converter=integer_converter)

    primitive = int

    def __int__(self):
        return self.value

    def __float__(self):
        return float(self.value)

    def __bool__(self):
        return bool(self.value)

    def __hash__(self):
        return hash(self.value)


@attrs.define(repr=False)
class Decimal(Singular, ScalarMixin):

    value: decimal.Decimal = attrs.field(converter=decimal_converter)

    primitive = float

    def __float__(self):
        return float(self.value)

    def __bool__(self):
        return bool(self.value)

    def __hash__(self):
        return hash(self.value)


@attrs.define(repr=False)
class Boolean(Singular, LogicalMixin):

    primitive = bool

    value: bool = attrs.field(converter=boolean_converter)

    def __str__(self):
        return str(self.value).lower()

    def __bool__(self):
        return self.value

    def __hash__(self):
        return hash(self.value)


@attrs.define(auto_attribs=False, repr=False)
class Array(WithClassifiers, Field):

    # WithClassifiers class attrs
    classifiers_attr_name: str = "item_type"
    multiple_classifiers: bool = False
    allowed_classifiers: ty.Tuple[ty.Type[Singular]] = (Singular,)
    item_type: ty.Union[ty.Type[Singular], None] = None

    primitive = tuple

    value: tuple = attrs.field(converter=array_converter)

    def __attrs_post_init__(self):
        # Ensure items are of the correct type
        if self.item_type is not None:
            self.value = tuple(self.item_type(i).value for i in self.value)

    def __str__(self):
        return (
            "["
            + ",".join(
                str(self.item_type(i))
                if self.item_type is not None
                else i  # pylint: disable=not-callable
                for i in self.value
            )
            + "]"
        )

    def __iter__(self):
        return iter(self.value)

    def __hash__(self):
        return hash(self.value)



from pathlib import Path
import json
from ..core import __version__, mark, DataType
from ..core.mixin import WithClassifiers
from ..generic import File
from ..core.exceptions import FormatMismatchError


class Schema(DataType):
    """Base class for all serialization schemas (can be used for abstract schemas that
    don't actually have a formal definition)"""

    pass


class JsonSchema(Schema):
    pass


class XmlSchema(Schema):
    pass


class InformalSchema(Schema):
    "Not actually a strict schema, just a set of conventions on how to structure the serialization"


class DataSerialization(WithClassifiers, File):
    "Base class for text-based hierarchical data-serialization formats, e.g. JSON, YAML"

    # Classifiers class attrs
    classifiers_attr_name = "schema"
    schema = None
    multiple_classifiers = False
    allowed_classifiers = (Schema,)
    generically_qualifies = True
    binary = False

    iana_mime = None

    @mark.extra
    def load(self):
        """Load the contents of the file into a dictionary"""
        raise NotImplementedError

    @mark.extra
    def save(data):
        """Serialise a dictionary to a new file"""
        raise NotImplementedError

    @classmethod
    def save_new(cls, fspath, data):
        # We have to use a mock object as the data file hasn't been written yet
        mock = cls.mock(fspath)
        mock.save(data)
        return cls(fspath)


class Xml(DataSerialization):
    ext = ".xml"
    allowed_classifiers = (XmlSchema, InformalSchema)


class Json(DataSerialization):
    ext = ".json"
    allowed_classifiers = (JsonSchema, InformalSchema)

    @mark.check
    def load(self):
        try:
            with open(self.fspath) as f:
                dct = json.load(f)
        except json.JSONDecodeError as e:
            raise FormatMismatchError(
                f"'{self.fspath}' is not a valid JSON file"
            ) from e
        return dct

    def save(self, data):
        with open(self.fspath, "w") as f:
            json.dump(data, f)


class Yaml(DataSerialization):
    ext = ".yaml"
    alternate_exts = (".yml",)


class Toml(DataSerialization):
    ext = ".toml"



from __future__ import annotations
import decimal
import pytest
import pydra.mark
from fileformats.core import from_mime, DataType, FileSet
from fileformats.core.mark import converter
from fileformats.archive import Zip
from fileformats.generic import DirectoryContaining
from fileformats.field import Array, Integer, Decimal, Text, Boolean
from fileformats.core.exceptions import (
    FileFormatsError,
    FormatConversionError,
    FormatRecognitionError,
    FormatMismatchError,
)
from fileformats.testing import (
    A,
    B,
    C,
    D,
    E,
    F,
    G,
    H,
    J,
    K,
    L,
    M,
    N,
    P,
    Q,
    R,
    TestField,
)


SpecificDataType = DataType.type_var("SpecificDataType")
SpecificFileSet = FileSet.type_var("SpecificFileSet")


def test_classified_equivalence():
    assert F is F
    assert F[A] is F[A]
    assert F[A, B] is F[B, A]
    assert K[A, B] is not K[B, A]  # ordered qualifers
    assert F[A] is not F[B]
    assert F[SpecificDataType] is F[SpecificDataType]
    assert F[SpecificDataType] is not F[A]
    assert F[SpecificDataType] is not F[SpecificFileSet]


def test_subtype_testing_1():
    assert issubclass(G, F)


def test_subtype_testing_2():
    assert not issubclass(F, G)


def test_subtype_testing_3():
    assert issubclass(J, J)


def test_subtype_testing_4():
    assert issubclass(J, H)


def test_subtype_testing_5():
    assert issubclass(J[A], J)


def test_subtype_testing_6():
    assert not issubclass(H[A], F)


def test_subtype_testing_7():
    assert not issubclass(H[A], F[A])


def test_subtype_testing_8():
    assert issubclass(J[A, B], J[A])


def test_subtype_testing_9():
    assert not issubclass(J[B], J[A])


def test_subtype_testing_10():
    assert not issubclass(J[A], J[A, B])


def test_subtype_testing_11():
    assert issubclass(J[A], H[A])


def test_subtype_testing_12():
    assert issubclass(J[A, B, C], H[A, B])


def test_subtype_testing_13():
    assert not issubclass(J[A], H[B])


def test_subtype_testing_14():
    assert issubclass(FileSet, DataType)


def test_subtype_testing_15():
    assert issubclass(SpecificFileSet, SpecificDataType)


def test_subtype_testing_15a():
    assert not issubclass(SpecificDataType, SpecificFileSet)


def test_subtype_testing_16():
    assert issubclass(F[SpecificFileSet], F[SpecificDataType])


def test_subtype_testing_16a():
    assert not issubclass(F[SpecificDataType], F[SpecificFileSet])


def test_subtype_testing_17():
    assert issubclass(L[A, E], L[A, C])


def test_subtype_testing_18():
    assert not issubclass(L[E, A], L[A, C])


def test_subtype_testing_19():
    assert issubclass(K[A, E, B], K[A, C, B])


def test_subtype_testing_20():
    assert issubclass(N[J, K], N[J, H])  # J is subclass of H,


def test_qualifier_fails():
    H[A, B, C]  # A, B, C are all allowable qualifier
    F[D]  # F has no restriction on qualifier types

    with pytest.raises(FileFormatsError) as e:
        H[D]
    assert "Invalid content types provided to" in str(e)

    with pytest.raises(FileFormatsError) as e:
        H[A, B, A]
    assert "Cannot have more than one occurrence of a qualifier" in str(e)

    K[A, B, A]  # ordered classifiers allow repeats

    with pytest.raises(FileFormatsError) as e:
        Q[A]
    assert (
        "Default value for classifiers attribute 'new_classifiers_attr' needs to be set"
        in str(e)
    )

    with pytest.raises(FileFormatsError) as e:
        M[A, B]
    assert "Multiple classifiers not permitted for " in str(e)


# (source_format=F, target_format=H)


@converter
@converter(source_format=F[A], target_format=H[A])  # Additional converter
@pydra.mark.task
def f2h(in_file: F) -> H:
    return in_file


def test_qualifier_converters():
    H.get_converter(F)
    assert F.get_converter(G) is None  # G is subtype of F
    with pytest.raises(FormatConversionError):  # Cannot convert to more specific type
        G.get_converter(F)
    H[A].get_converter(F[A])
    with pytest.raises(FormatConversionError):
        H[B].get_converter(F[B])
    assert F[A].get_converter(G[A]) is None
    assert F.get_converter(G[A]) is None
    with pytest.raises(FormatConversionError):
        F[A].get_converter(F[E])
    with pytest.raises(FormatConversionError):
        F[A].get_converter(G)


@converter(source_format=K[A, B], target_format=L[A, B])
@converter(source_format=K[C, D], target_format=L[D, C])
@converter(source_format=K[A, B, C], target_format=L[A, B])
@converter(source_format=K[A, C, B], target_format=L[C, A])
@pydra.mark.task
def k2l(in_file: K) -> L:
    return in_file


def test_ordered_qualifier_converters():
    L[A, B].get_converter(K[A, B])
    with pytest.raises(FormatConversionError):
        L[B, A].get_converter(K[A, B])
    L[D, C].get_converter(K[C, D])
    with pytest.raises(FormatConversionError):
        L[C, D].get_converter(K[C, D])

    L[C, A].get_converter(K[A, C, B])
    L[C, A].get_converter(K[A, E, B])  # E is a subtype of C


def test_mime_roundtrips():
    assert DirectoryContaining[F].mime_like == "testing/f+directory-containing"
    assert from_mime("testing/f+directory-containing") is DirectoryContaining[F]

    # Directory is unordered so sort classifiers to create unique mime
    assert DirectoryContaining[H, F].mime_like == "testing/f.h+directory-containing"
    assert from_mime("testing/f.h+directory-containing") is DirectoryContaining[F, H]

    # K is ordered
    assert K[B, A].mime_like == "testing/b.a+k"
    assert from_mime("testing/b.a+k") is K[B, A]
    assert from_mime("testing/b.a+k") is not K[A, B]

    with pytest.raises(FormatRecognitionError) as e:
        Array[TestField].mime_like
    assert "Cannot create reversible MIME type for " in str(e)


def test_inherited_classifiers():
    assert Zip[G].mime_like == "testing/g+zip"
    assert from_mime("testing/g+zip") is Zip[G]


def test_arrays():
    Array[Integer]([1, 2, 3, 4])
    with pytest.raises(FormatMismatchError) as e:
        Array[Integer]([1.5, 2.2])
    assert "Cannot convert float (1.5) to integer field" in str(e)

    assert list(Array[Decimal](["1.5", "2.2"])) == [
        decimal.Decimal("1.5"),
        decimal.Decimal("2.2"),
    ]
    assert list(Array[Decimal](["1.5", "2.2"])) == [
        decimal.Decimal("1.5"),
        decimal.Decimal("2.2"),
    ]
    assert list(Array[Decimal]("1.5, 2.2")) == [
        decimal.Decimal("1.5"),
        decimal.Decimal("2.2"),
    ]
    assert list(Array[Decimal]("[1.5, 2.2]")) == [
        decimal.Decimal("1.5"),
        decimal.Decimal("2.2"),
    ]
    assert list(Array[Text]("[1.5, 2.2]")) == ["1.5", "2.2"]

    assert list(Array[Boolean]("yes, YES, no, 0, 1, False, True, true")) == [
        True,
        True,
        False,
        False,
        True,
        False,
        True,
        True,
    ]

    assert list(from_mime("field/integer+array")("1,2,3,4,5")) == [1, 2, 3, 4, 5]


# Template-from-template conversions


@converter
@pydra.mark.task
def f2n_template(in_file: F[SpecificDataType]) -> N[SpecificDataType]:
    return in_file


@converter
@pydra.mark.task
def f2p_template(in_file: F[SpecificDataType]) -> P[SpecificDataType]:
    return in_file


@converter
@pydra.mark.task
def p2n_template(in_file: P[SpecificDataType]) -> N[SpecificDataType]:
    return in_file


def test_wildcard_template_from_template_conversion():
    H[A].get_converter(F[A])
    with pytest.raises(FormatConversionError):
        H[B].get_converter(F[A])


# Template from generic type to template


@converter
@pydra.mark.task
def generic2f(in_file: SpecificDataType) -> F[SpecificDataType]:
    return in_file


@converter
@pydra.mark.task
def generic2n(in_file: SpecificDataType) -> N[SpecificDataType, H]:
    return in_file


def test_wildcard_template_from_generic_conversion():
    F[J].get_converter(J)
    with pytest.raises(FormatConversionError):
        F[K].get_converter(J)

    N[J].get_converter(J)
    N[J, H].get_converter(J)
    with pytest.raises(FormatConversionError):
        F[K, H].get_converter(J)
    with pytest.raises(FormatConversionError):
        F[J, K].get_converter(J)


# Generic from template to  type


@converter
@pydra.mark.task
def f2generic(in_file: F[SpecificDataType]) -> SpecificDataType:
    return in_file


@converter
@pydra.mark.task
def n2generic(in_file: N[SpecificDataType, G]) -> SpecificDataType:
    return in_file


def test_wildcard_generic_from_template_conversion():
    J.get_converter(F[J])
    with pytest.raises(FormatConversionError):
        J.get_converter(F[K])


def test_wildcard_generic_from_multi_template_conversion():
    J.get_converter(N[J, G])
    with pytest.raises(FormatConversionError):
        J.get_converter(N[J, K])


@converter
@pydra.mark.task
def l2r(in_file: L[A, SpecificDataType, C]) -> R[A, SpecificDataType, C, D]:
    return in_file


def test_wildcard_ordered_qualifier_converters():
    R[A, B, C, D].get_converter(L[A, B, C])
    with pytest.raises(FormatConversionError):
        R[A, E, C, D].get_converter(L[A, B, C])



from abc import ABCMeta
import typing as ty
import attrs
from .utils import describe_task
from .exceptions import FileFormatsError

if ty.TYPE_CHECKING:
    from .datatype import DataType


@attrs.define
class ConverterWrapper:
    """Wraps a converter task in a workflow so that the in_file and out_file names can
    be mapped onto their standardised names, "in_file" and "out_file" if necessary
    """

    task_spec: ty.Callable
    in_file: ty.Optional[str] = None
    out_file: ty.Optional[str] = None

    def __call__(self, name=None, **kwargs):
        from pydra.engine import Workflow

        if name is None:
            name = f"{self.task_spec.__name__}_wrapper"
        wf = Workflow(
            name=name, input_spec=list(set(["in_file"] + list(kwargs))), **kwargs
        )
        wf.add(self.task_spec(name="task", **{self.in_file: wf.lzin.in_file}))
        wf.set_output([("out_file", getattr(wf.task.lzout, self.out_file))])
        return wf


class SubtypeVar:
    """To handle the case where the target format is a placeholder (type-var) defined by
    by its relationship to the source format, e.g.

    AnyFileFormat = FileSet.type_var("AnyFileFormat")

    @converter
    @pydra.mark.task
    def unzip(in_file: Zip[AnyFileFormat], out_file: AnyFileFormat):
        ...
    """

    converters = {}

    @classmethod
    def new(cls, name: str, klass: ty.Type["DataType"]) -> "SubtypeVar":
        """Create a new subtype

        Parameters
        ----------
        name : str
            name for the subtype
        klass : ty.Type[DataType]
            the class to sub-type

        Returns
        -------
        SubtypeVar
            a sub-type that is
        """
        return ABCMeta(name, (cls, klass), {"bound": klass})

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        if issubclass(subclass, SubtypeVar):
            return issubclass(subclass.bound, cls.bound)
        return type.__subclasscheck__(cls, subclass)

    @classmethod
    def get_converter_tuples(
        cls, source_format: type, target_format: type
    ) -> ty.List[ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]]]:
        # check to see whether there are converters from a base class of the source
        # format
        available_converters = []
        if source_format.is_classified:
            for template_source_format, converter in cls.converters.items():
                if not issubclass(
                    template_source_format.unclassified, source_format.unclassified
                ):
                    continue
                assert len(template_source_format.wildcard_classifiers()) == 1
                non_wildcards = template_source_format.non_wildcard_classifiers()
                if not non_wildcards.issubset(source_format.classifiers):
                    continue
                from_types = tuple(
                    set(source_format.classifiers).difference(non_wildcards)
                )
                if any(issubclass(q, target_format) for q in from_types):
                    available_converters.append(converter)
        return available_converters

    @classmethod
    def register_converter(
        cls,
        source_format: type,
        converter_tuple: ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]],
    ):
        """Registers a converter task within a class attribute. Called by the @fileformats.mark.converter
        decorator.

        Parameters
        ----------
        source_format : type
            the source format to register a converter from
        task_spec : ty.Callable
            a callable that resolves to a Pydra task
        converter_kwargs : dict
            additional keyword arguments to be passed to the task spec at initialisation
            time

        Raises
        ------
        FormatConversionError
            if there is already a converter registered between the two types
        """
        # Ensure "converters" dict is defined in the target class and not in a superclass
        if len(source_format.wildcard_classifiers()) > 1:
            raise FileFormatsError(
                "Cannot register a conversion to a generic type from a type with more "
                f"than one wildcard {source_format} ({list(source_format.wildcard_classifiers())})"
            )
        prev_registered = [
            f
            for f in cls.converters
            if (
                f.unclassified is source_format.unclassified
                and f.non_wildcard_classifiers()
                == source_format.non_wildcard_classifiers()
            )
        ]
        assert len(prev_registered) <= 1
        if prev_registered:
            prev = prev_registered[0]
            prev_task = cls.converters[prev][0]
            raise FileFormatsError(
                f"There is already a converter registered from {prev} "
                f"to the generic type '{tuple(prev.wildcard_classifiers())[0]}':"
                f"{describe_task(prev_task)}"
            )

        cls.converters[source_format] = converter_tuple



import attrs
import pytest
from pydra.engine.specs import File
from fileformats.testing import Foo, Bar, Baz, Qux
from fileformats.core import mark
from fileformats.core.exceptions import FormatConversionError
from conftest import write_test_file

try:
    import pydra.mark
except ImportError:
    pydra = None


@pytest.fixture
def foo_bar_converter(work_dir):
    @mark.converter
    @pydra.mark.task
    @pydra.mark.annotate({"return": {"out_file": Bar}})
    def foo_bar_converter_(in_file: Foo):
        return Bar(write_test_file(work_dir / "bar.bar", in_file.contents))

    return foo_bar_converter_


@pytest.fixture
def baz_bar_converter(work_dir):
    @mark.converter(out_file="out")
    @pydra.mark.task
    @pydra.mark.annotate({"return": {"out": Bar}})
    def baz_bar_converter_(in_file: Baz):
        assert in_file
        return Bar(write_test_file(work_dir / "bar.bar", in_file.contents))

    return baz_bar_converter_


@pytest.fixture
def FooQuxConverter():
    from pydra.engine import specs
    from pydra import ShellCommandTask

    input_fields = [
        (
            "in_file",
            File,
            {
                "help_string": "the input file",
                "argstr": "",
            },
        ),
        (
            "out_file",
            str,
            {
                "help_string": "output file name",
                "argstr": "",
                "position": -1,
                "output_file_template": "out.qux",
            },
        ),
    ]

    FooQux_input_spec = specs.SpecInfo(
        name="Input", fields=input_fields, bases=(specs.ShellSpec,)
    )

    output_fields = [
        (
            "out_file",
            File,
            {
                "help_string": "output file",
            },
        ),
    ]
    FooQux_output_spec = specs.SpecInfo(
        name="Output", fields=output_fields, bases=(specs.ShellOutSpec,)
    )

    @mark.converter(source_format=Foo, target_format=Qux)
    class FooQuxConverter_(ShellCommandTask):

        input_spec = FooQux_input_spec
        output_spec = FooQux_output_spec
        executable = "cp"

    return FooQuxConverter_


@pytest.mark.skipif(pydra is None, reason="Pydra could not be imported")
def test_get_converter_functask(foo_bar_converter, work_dir):

    fspath = work_dir / "test.foo"
    write_test_file(fspath)
    assert attrs.asdict(Bar.get_converter(Foo, name="Foo2Bar").inputs) == attrs.asdict(
        foo_bar_converter(name="Foo2Bar").inputs
    )


@pytest.mark.skipif(pydra is None, reason="Pydra could not be imported")
def test_get_converter_shellcmd(FooQuxConverter, work_dir):

    fspath = work_dir / "test.foo"
    write_test_file(fspath)
    assert attrs.asdict(Qux.get_converter(Foo, name="Foo2Qux").inputs) == attrs.asdict(
        FooQuxConverter(name="Foo2Qux").inputs
    )


@pytest.mark.skipif(pydra is None, reason="Pydra could not be imported")
def test_get_converter_fail(work_dir):

    fspath = work_dir / "test.foo"
    write_test_file(fspath)
    with pytest.raises(FormatConversionError):
        Baz.get_converter(Foo, name="Foo2Baz")


@pytest.mark.skipif(pydra is None, reason="Pydra could not be imported")
def test_convert_functask(work_dir):

    fspath = work_dir / "test.foo"
    write_test_file(fspath)
    foo = Foo(fspath)
    bar = Bar.convert(foo)
    assert type(bar) is Bar
    assert bar.contents == foo.contents


@pytest.mark.skipif(pydra is None, reason="Pydra could not be imported")
def test_convert_shellcmd(work_dir):

    fspath = work_dir / "test.foo"
    write_test_file(fspath)
    foo = Foo(fspath)
    qux = Qux.convert(foo)
    assert type(qux) is Qux
    assert qux.contents == foo.contents


@pytest.mark.skipif(pydra is None, reason="Pydra could not be imported")
def test_convert_mapped_conversion(baz_bar_converter, work_dir):

    fspath = work_dir / "test.baz"
    write_test_file(fspath)
    baz = Baz(fspath)
    bar = Bar.convert(baz)
    assert type(bar) is Bar
    assert bar.contents == baz.contents



from __future__ import annotations
import importlib
from pathlib import Path
import inspect
import typing as ty
import re
import urllib.request
import urllib.error
import os
import logging
import pkgutil
from contextlib import contextmanager
from fileformats.core.exceptions import (
    FileFormatsError,
)
import fileformats.core

logger = logging.getLogger("fileformats")


_excluded_subpackages = set(["core", "testing"])


def include_testing_package(flag: bool = True):
    """Include testing package in list of sub-packages. Typically set in conftest.py
    or similar when setting up unittesting. Must be set globally before any methods are
    called within the package as member classes are cached.

    Parameters
    ----------
    flag : bool
        whether to include the testing package or not
    """
    global _excluded_subpackages
    if flag:
        _excluded_subpackages.remove("testing")
    else:
        _excluded_subpackages.add("testing")


def find_matching(
    fspaths: ty.List[Path],
    candidates: ty.Sequence = None,
    standard_only: bool = False,
    include_generic: bool = False,
    skip_unconstrained: bool = True,
):
    """Detect the corresponding file format from a set of file-system paths

    Parameters
    ----------
    fspaths : list[Path]
        file-system paths to detect the format of
    candidates: sequence[DataType], optional
        the candidates to select from, by default all file formats
    standard_only : bool, optional
        If you only want to return matches from the "standard" IANA types, by default False
    skip_unconstrained : bool, optional
        skip formats that aren't constrained by extension, magic number or another check
    """
    import fileformats.core.mixin

    fspaths = fspaths_converter(fspaths)
    matches = []
    if candidates is None:
        candidates = fileformats.core.FileSet.all_formats
    for frmt in candidates:
        if skip_unconstrained and frmt.unconstrained:
            continue
        namespace = frmt.namespace
        if (
            frmt.matches(fspaths)
            and (not standard_only or namespace in STANDARD_NAMESPACES)
            and (include_generic or namespace != "generic")
        ):
            matches.append(frmt)
    return matches


def from_mime(mime_str: str):
    """Resolves a MIME type (or MIME-like) string into the corresponding type

    Parameters
    ----------
    mime_str : str
        the MIME type, or MIME-like (i.e. using the fileformats namespace scheme
        instead of putting all non-standard types into application/*), string to
        resolve

    Returns
    -------
    datatype : type
        the resolved datatype
    """
    if mime_str.endswith(LIST_MIME):
        item_mime = mime_str[: -len(LIST_MIME)]
        if item_mime.startswith("[") and item_mime.endswith("]"):
            item_mime = item_mime[1:-1]
        return ty.List[from_mime(item_mime)]
    if "," in mime_str:
        return ty.Union.__getitem__(tuple(from_mime(t) for t in mime_str.split(",")))
    return fileformats.core.DataType.from_mime(mime_str)


def to_mime(datatype: type, official=False):
    """Returns the mime-type or mime-like (i.e. using fileformats namespaces instead
    of putting all non-standard types in the applications/* registry) string corresponding
    to the given datatype

    Parameters
    ----------
    datatype : type
        the datatype to get the mime string for
    official : bool
        whether to use the official mime-type instead of mime-like

    Returns
    -------
    mime_str : str
        the MIME type string if `iana=True`, or MIME-like (i.e. using the fileformats
        namespace scheme instead of putting all non-standard types into application/*)
        if not
    """
    origin = ty.get_origin(datatype)
    if official and (origin or datatype.namespace == "field"):
        raise TypeError(
            f"Cannot convert {datatype} to official mime-type as it is not a proper "
            'file-type, please use official=False to convert to "mime-like" string instead'
        )
    if origin is list:
        item_mime = to_mime(ty.get_args(datatype)[0])
        if "," in item_mime:
            item_mime = "[" + item_mime + "]"
        item_mime += LIST_MIME
        return item_mime
    if origin is ty.Union:
        return ",".join(t.mime_like for t in ty.get_args(datatype))
    return datatype.mime_type if official else datatype.mime_like


def subpackages(exclude: ty.Sequence[str] = _excluded_subpackages):
    """Iterates over all subpackages within the fileformats namespace

    Parameters
    ----------
    exclude : ty.Sequence[str], optional
        whether to include the testing subpackage, by default ["core", "testing"]

    Yields
    ------
    module
        all modules within the package
    """
    for mod_info in pkgutil.iter_modules(
        fileformats.__path__, prefix=fileformats.__package__ + "."
    ):
        if mod_info.name.split(".")[-1] in exclude:
            continue
        yield importlib.import_module(mod_info.name)


@contextmanager
def set_cwd(path: Path):
    """Sets the current working directory to `path` and back to original
    working directory on exit

    Parameters
    ----------
    path : str
        The file system path to set as the current working directory
    """
    pwd = os.getcwd()
    os.chdir(path)
    try:
        yield path
    finally:
        os.chdir(pwd)


def fspaths_converter(
    fspaths: ty.Union[
        ty.Iterable[ty.Union[str, os.PathLike, bytes]],
        str,
        os.PathLike,
        bytes,
        fileformats.core.FileSet,
    ]
):
    """Ensures fs-paths are a set of pathlib.Path"""
    import fileformats.core

    if isinstance(fspaths, fileformats.core.FileSet):
        fspaths = fspaths.fspaths
    elif isinstance(fspaths, (str, os.PathLike, bytes)):
        fspaths = [fspaths]
    return frozenset(Path(p).absolute() for p in fspaths)


class classproperty(object):
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


STANDARD_NAMESPACES = [
    "archive",
    "audio",
    "document",
    "image",
    "misc",
    "model",
    "numeric",
    "serialization",
    "text",
    "video",
]


def to_mime_format_name(format_name: str):
    if "___" in format_name:
        raise FileFormatsError(
            f"Cannot convert name of format class {format_name} to mime string as it "
            "contains triple underscore"
        )
    if format_name.startswith("_"):
        format_name = format_name[1:]
    format_name = format_name[0].lower() + format_name[1:]
    format_name = re.sub("__([A-Z])", lambda m: "+" + m.group(1).lower(), format_name)
    format_name = re.sub("_([A-Z])", lambda m: "." + m.group(1).lower(), format_name)
    format_name = re.sub("([A-Z])", lambda m: "-" + m.group(1).lower(), format_name)
    return format_name


def from_mime_format_name(format_name: str):
    if format_name.startswith("x-"):
        format_name = format_name[2:]
    if re.match(r"^[0-9]", format_name):
        format_name = "_" + format_name
    format_name = format_name.capitalize()
    format_name = re.sub(r"(\.)(\w)", lambda m: "_" + m.group(2).upper(), format_name)
    format_name = re.sub(r"(\+)(\w)", lambda m: "__" + m.group(2).upper(), format_name)
    format_name = re.sub(r"(-)(\w)", lambda m: m.group(2).upper(), format_name)
    return format_name


def add_exc_note(e, note):
    """Adds a note to an exception in a Python <3.11 compatible way

    Parameters
    ----------
    e : Exception
        the exception to add the note to
    note : str
        the note to add

    Returns
    -------
    Exception
        returns the exception again
    """
    if hasattr(e, "add_note"):
        e.add_note(note)
    else:
        e.args = (e.args[0] + "\n" + note,)
    return e


def describe_task(task):
    """Returns the name of a Pydra task and where it was defined for debugging purposes

    Parameters
    ----------
    task : pydra.engine.core.TaskBase
        the task to describe
    """
    from fileformats.core.converter import ConverterWrapper

    if isinstance(task, ConverterWrapper):
        task = task.task_spec
    if inspect.isfunction(task):
        import cloudpickle

        task = cloudpickle.loads(task().inputs._func)
    src_file = inspect.getsourcefile(task)
    src_line = inspect.getsourcelines(task)[-1]
    return f"{task} (defined at line {src_line} of {src_file})"


def check_package_exists_on_pypi(package_name: str, timeout: int = 5) -> bool:
    """Check if a package exists on PyPI

    Parameters
    ----------
    package_name : str
        the name of the package to check for

    Returns
    -------
    bool
        whether the package exists on PyPI or not
    """
    url = f"https://pypi.org/pypi/{package_name}/json"
    try:
        urllib.request.urlopen(url, timeout=timeout)
    except urllib.error.HTTPError as e:
        if e.code == 404:
            return False
        else:
            raise
    return True


def import_extras_module(klass: type) -> ty.Tuple[bool, str]:
    """Attempt to load extras module corresponding to the provided class's module

    Parameters
    ----------
    klass : ty.Type
        the class to load the extras module for

    Returns
    -------
    imported : bool
        whether the module was imported or not
    sub_pkg : str
        the name of the sub-package that was attempted to be loaded
    """
    # Check for Mock class
    try:
        klass = klass.TRUE_CLASS
    except AttributeError:
        pass
    pkg_parts = klass.__module__.split(".")
    if pkg_parts[0] != "fileformats":
        logger.debug(
            "There is no 'extras' module for classes not within the 'fileformats' package, "
            "not %s in %s",
            klass.__name__,
            klass.__module__,
        )
        return True, None, None
    sub_pkg = pkg_parts[1]
    extras_pkg = "fileformats.extras." + sub_pkg
    if sub_pkg in STANDARD_NAMESPACES:
        extras_pypi = "fileformats-extras"
    else:
        extras_pypi = f"fileformats-{sub_pkg}-extras"
    try:
        importlib.import_module(extras_pkg)
    except ImportError:
        extras_imported = False
    else:
        extras_imported = True
    return extras_imported, extras_pkg, extras_pypi


LIST_MIME = "+list-of"



from ..core import __version__
from .basic import Foo, Bar, Baz, Qux
from .headers import (
    Y,
    Xyz,
    MyFormat,
    MyFormatGz,
    MyFormatGzX,
    MyFormatX,
    YourFormat,
    SeparateHeader,
    ImageWithHeader,
    EncodedText,
)
from .classifiers import A, B, C, D, E, F, G, H, J, K, L, M, N, P, Q, R, TestField



import os
import typing as ty
from pathlib import Path
import attrs
from fileformats.core.fileset import FileSet
from fileformats.core.exceptions import FormatMismatchError, FileFormatsError
from fileformats.core import mark
from fileformats.core.utils import classproperty
from fileformats.core.mixin import WithClassifiers


class FsObject(FileSet, os.PathLike):
    "Generic file-system object, can be either a file or a directory"

    @mark.required
    @property
    def fspath(self):
        if len(self.fspaths) > 1:
            raise FormatMismatchError(
                f"More than one fspath ({self.fspaths}) provided to {self}, "
                f"primary path is ambiguous"
            )
        return next(iter(self.fspaths))

    def __str__(self):
        return str(self.fspath)

    def __fspath__(self):
        """Render to string, so can be treated as any other file-system path, i.e. passed
        to functions like file 'open'"""
        return str(self)

    @property
    def stem(self):
        return self.fspath.with_suffix("").name

    @classproperty
    def unconstrained(cls) -> bool:
        """Whether the file-format is unconstrained by extension, magic number or another
        constraint"""
        # We have to subtract `fspath` from required properties as we defined unconstrained
        # file-sets as ones that have more constraints than simply existing
        return not (len(list(cls.required_properties())) - 1)


class File(FsObject):
    """Generic file type"""

    binary = True
    is_dir = False

    @mark.required
    @property
    def fspath(self):
        fspath = self.select_by_ext()
        if fspath.is_dir():
            # fspath is guaranteed to exist
            raise FormatMismatchError(
                f'Path that matches extension of {type(self)}, "{fspath}", '
                f"is a directory not a file"
            )
        return fspath

    @classproperty
    def unconstrained(cls) -> bool:
        """Whether the file-format is unconstrained by extension, magic number or another
        constraint"""
        return super().unconstrained and (cls.ext is None or None in cls.alternate_exts)

    @classmethod
    def copy_ext(
        cls,
        old_path: Path,
        new_path: Path,
        decomposition_mode=FileSet.ExtensionDecomposition.none,
    ):
        """Copy extension from the old path to the new path, ensuring that all
        of the extension is used (e.g. 'my.gz' instead of 'gz')

        Parameters
        ----------
        old_path: Path or str
            The path from which to copy the extension from
        new_path: Path or str
            The path to append the extension to
        decomposition_mode : FileSet.ExtensionDecomposition, optional
            if the file doesn't have an explicit extension, how to interpret "." within
            the filename

        Returns
        -------
        Path
            The new path with the copied extension
        """
        if not cls.matching_exts([old_path], [cls.ext]):
            raise FormatMismatchError(
                f"Extension of old path ('{str(old_path)}') does not match that "
                f"of file, '{cls.ext}'"
            )
        suffix = (
            cls.ext
            if cls.ext
            else cls.decompose_fspath(old_path, mode=decomposition_mode)[-1]
        )
        return Path(new_path).with_suffix(suffix)

    @property
    def contents(self):
        return self.read_contents()

    def read_contents(self, size=None, offset=0):
        with open(self.fspath, "rb" if self.binary else "r") as f:
            if offset:
                f.read(offset)
            contents = f.read(size)
        return contents

    @property
    def actual_ext(self):
        "The actual file extension (out of the primary  and alternate extensions possible)"
        matching = [e for e in self.possible_exts if self.fspath.name.endswith(e)]
        if not matching:
            assert False, (
                f"extension of fspath {self.fspath} is not in possible extensions for "
                f"{type(self)} class: {self.possible_exts}. This should have been "
                "checked at initialisation"
            )
        # Return the longest matching extension, useful for optional extensions
        return sorted(matching, key=len)[-1]

    @property
    def stem(self):
        if self.actual_ext:
            stem = self.fspath.name[: -len(self.actual_ext)]
        else:
            stem = self.fspath
        return stem


class Directory(FsObject):
    """Base directory to be overridden by subtypes that represent directories but don't
    want to inherit content type "qualifers" (i.e. most of them)"""

    is_dir = True

    content_types = ()

    @mark.required
    @property
    def fspath(self):
        # fspaths are checked for existence with the exception of mock classes
        dirs = [p for p in self.fspaths if not p.is_file()]
        if not dirs:
            raise FormatMismatchError(f"No directory paths provided {repr(self)}")
        if len(dirs) > 1:
            raise FormatMismatchError(
                f"More than one directory path provided {dirs} to {repr(self)}"
            )
        fspath = dirs[0]
        missing = []
        for content_type in self.content_types:
            match = False
            for p in fspath.iterdir():
                try:
                    content_type([p])
                except FormatMismatchError:
                    continue
                else:
                    match = True
                    break
            if not match:
                missing.append(content_type)
        if missing:
            raise FormatMismatchError(
                f"Did not find matches for {missing} content types in {repr(self)}"
            )
        return fspath

    @property
    def contents(self):
        for content_type in self.content_types:
            for p in self.fspath.iterdir():
                try:
                    yield content_type([p])
                except FormatMismatchError:
                    continue

    @classproperty
    def unconstrained(cls) -> bool:
        """Whether the file-format is unconstrained by extension, magic number or another
        constraint"""
        return super().unconstrained and not cls.content_types

    @mark.check
    def validate_contents(self):
        if not self.content_types:
            return
        not_found = set(self.content_types)
        for fspath in self.fspath.iterdir():
            for content_type in list(not_found):
                if content_type.matches(fspath):
                    not_found.remove(content_type)
                    if not not_found:
                        return
        assert not_found
        raise FormatMismatchError(
            f"Did not find the required content types, {not_found}, within the "
            f"directory {self.fspath} of {self}"
        )

    def hash_files(self, relative_to=None, **kwargs):
        if relative_to is None:
            relative_to = self.fspath
        return super().hash_files(relative_to=relative_to, **kwargs)


class TypedSet(FileSet):
    """List of specific file types (similar to the contents of a directory but not
    enclosed in one)"""

    content_types = ()

    @property
    def contents(self):
        for content_type in self.content_types:
            for p in self.fspaths:
                try:
                    yield content_type([p])
                except FormatMismatchError:
                    continue

    @mark.check
    def validate_contents(self):
        if not self.content_types:
            return
        not_found = set(self.content_types)
        for fspath in self.fspaths:
            for content_type in list(not_found):
                if content_type.matches(fspath):
                    not_found.remove(content_type)
                    if not not_found:
                        return
        assert not_found
        raise FormatMismatchError(
            f"Did not find the required content types, {not_found}, within the "
            f"given list {self.fspaths}"
        )


class DirectoryContaining(WithClassifiers, Directory):
    """Generic directory classified by the formats of its contents"""

    # WithClassifiers-required class attrs
    classifiers_attr_name = "content_types"
    allowed_classifiers = (FileSet,)
    generically_qualifies = True


class SetOf(WithClassifiers, TypedSet):
    # WithClassifiers-required class attrs
    classifiers_attr_name = "content_types"
    allowed_classifiers = (FileSet,)
    generically_qualifies = True



from pathlib import Path
import typing as ty
from collections import Counter
from . import mark
from .fileset import FileSet
from .utils import classproperty, describe_task, to_mime_format_name
from .converter import SubtypeVar
from .exceptions import FileFormatsError, FormatMismatchError, FormatRecognitionError


class WithMagicNumber:
    """Mixin class for Files with magic numbers at the start of their
    contents.

    Class Attrs
    -----------
    magic_number : str
        the magic number/string to search for at the start of the file
    binary : bool
        if the file-format is a binary type then this flag needs to be set in order to
        read the contents properly
    magic_number_offset : int, optional
        the offset in bytes from the start of the file that the magic number is stored
    """

    magic_number_offset = 0
    binary: bool
    magic_number: ty.Union[str, int]

    @mark.check
    def check_magic_number(self):
        if self.binary and isinstance(self.magic_number, str):
            magic_bytes = bytes.fromhex(self.magic_number)
        else:
            magic_bytes = self.magic_number
        read_magic_number = self.read_contents(
            len(magic_bytes), offset=self.magic_number_offset
        )
        if read_magic_number != magic_bytes:
            if self.binary and isinstance(self.magic_number, str):
                read_magic_str = '"' + bytes.hex(read_magic_number) + '"'
                magic_str = '"' + self.magic_number + '"'
            else:
                read_magic_str = read_magic_number
                magic_str = self.magic_number
            raise FormatMismatchError(
                f"Magic number of file {read_magic_str} doesn't match expected "
                f"{magic_str}"
            )


class WithAdjacentFiles:
    """
    If only the main fspath is provided to the __init__ of the class, this mixin
    automatically includes any "adjacent files", i.e. any files with the same stem but
    different extensions

    Note that WithAdjacentFiles must come before the primary type in the method-resolution
    order of the class so it can override the '__attrs_post_init__' method in
    post_init_super class (typically FileSet), e.g.

        class MyFileFormatWithSeparateHeader(WithSeparateHeader, MyFileFormat):

            header_type = MyHeaderType

    Class Attrs
    -----------
    post_init_super : type
        the format class the WithAdjacentFiles mixin is mixed with that defines the
        __attrs_post_init__ method that should be called once the adjacent files
        are added to the self.fspaths attribute to run checks.
    """

    post_init_super = FileSet
    fspaths: ty.FrozenSet[Path]

    def __attrs_post_init__(self):
        if len(self.fspaths) == 1:
            self.fspaths |= self.get_adjacent_files()
            trim = True
        else:
            trim = False
        self.post_init_super.__attrs_post_init__(self)
        if trim:
            self.trim_paths()

    def get_adjacent_files(self) -> ty.Set[Path]:
        stem = self.stem  # pylint: disable=no-member
        adjacents = set()
        for sibling in self.fspath.parent.iterdir():  # pylint: disable=no-member
            if (
                sibling != self.fspath  # pylint: disable=no-member
                and sibling.is_file()
                and sibling.name.startswith(stem + ".")
            ):
                adjacents.add(sibling)
        return adjacents


class WithSeparateHeader(WithAdjacentFiles):
    """Mixin class for Files with metadata stored in separate header files (typically
    with the same file stem but differing extension)

    Note that WithSeparateHeader must come before the primary type in the method-resolution
    order of the class so it can override the '__attrs_post_init__' method, e.g.

        class MyFileFormatWithSeparateHeader(WithSeparateHeader, MyFileFormat):

            header_type = MyHeaderType

    Class Attrs
    -----------
    header_type : type
        the file-format of the header file
    """

    @mark.required
    @property
    def header(self):
        return self.header_type(self.select_by_ext(self.header_type))

    def read_metadata(self):
        return self.header.load()


class WithSideCars(WithAdjacentFiles):
    """Mixin class for Files with a "side-car" file that augments the inline metadata
    (typically with the same file stem but differing extension).

    Note that WithSideCars must come before the primary type in the method-resolution
    order of the class so it can override the '__attrs_post_init__' and 'read_metadata'
    methods, e.g.

        class MyFileFormatWithSideCars(WithSideCars, MyFileFormat):

            primary_type = MyFileFormat
            side_car_types = (MySideCarType,)

    Class Attrs
    -----------
    primary_type : type
        the file-format of the primary file (used to read the inline metadata), can be
        the base class that implements 'read_metadata'
    side_car_types : tuple[type, ...]
        the file-formats of the expected side-car files
    """

    @mark.required
    @property
    def side_cars(self):
        return [tp(self.select_by_ext(tp)) for tp in self.side_car_types]

    def read_metadata(self):
        metadata = self.primary_type.read_metadata(self)
        for side_car in self.side_cars:
            try:
                side_car_metadata = side_car.load()
            except AttributeError:
                continue
            else:
                metadata[
                    to_mime_format_name(type(side_car).__name__)
                ] = side_car_metadata
        return metadata


class WithClassifiers:
    """Mixin class for adding the ability to qualify the format class to designate the
    type of information stored within the format, e.g. ``DirectoryContaining[Png, Gif]`` for a
    directory containing PNG and GIF files, ``Zip[DataFile]`` for a zipped data file,
    ``Array[Integer]`` for an array containing integers, or DicomDir[T1w, Brain] for a
    T1-weighted MRI scan of the brain in DICOM format.

        class MyFormatWithContents(WithContents, File):

            ext = ".myf


        def my_func(file: MyFormatWithContents[Integer]):
            ...

    A unique class will be returned (i.e. multiple calls with the same arguments will
    return the same class)

    Class Attrs
    -----------
    classifiers_attr_name : str, optional
        an attribute name to store the classifiers at within the classified class. This
        should be used if you need to reference the ``classifiers`` attribute directly
        in any validation/other methods (i.e. in most cases), to handle the case of
        diamond inheritance between two classes that can be classified.
        A default value should also be set in the unclassified base for this attribute,
        which is either ``()`` if ``multiple_classifiers`` is true and ``None`` otherwise
    <classifiers-attr-name> : tuple[type, ...] or None, optional
        pass a default value to the attribute referenced by 'classifiers_attr_name'
    multiple_classifiers : bool, optional
        whether or not multiple content types are permitted for the container type, True
    allowed_classifiers : tuple[type,...], optional
        the allowable types (+ subclasses) for the content types. If None all types
        are allowed
    ordered_classifiers : bool, optional
        whether the order of the content types is important or not, by default false
    genericly_classified : bool, optional
        whether the class can be classified by classifiers in any namespace (true) or just the
        namespace it belongs to (false). If true, then the namespace of the genericly
        classified class is omitted from the "mime-like" string. Note that the
        class' name therefore needs to be globally unique amongst all other genericly
        classified classes and so it should be used sparingly, i.e., highly generic
        formats that are unambiguous across all namespaces, such as "directory", "zip",
        "gzip", "json", "yaml", etc...
    """

    classifiers = ()  # classifiers set in the current class
    _classified_subtypes = {}
    # dict of previously created classified subtypes. If an existing class with matching
    # classifiers has been created it is returned instead of creating a new type. This
    # ensures that ``assert MyFormat[Qualifier] is MyFormat[Qualifier]``

    # Default values for class attrs
    multiple_classifiers = True
    allowed_classifiers: ty.Optional[ty.Tuple[ty.Type[ty.Any]]] = None
    ordered_classifiers = False
    generically_qualifies = False

    def __attrs_pre_init__(self):
        if self.wildcard_classifiers():
            raise FileFormatsError(
                f"Can instantiate {type(self)} class as it has wildcard classifiers "
                "and therefore should only be used for converter specifications"
            )

    @classproperty
    def is_classified(cls):  # pylint: disable=no-self-argument
        return "unclassified" in cls.__dict__

    @classmethod
    def wildcard_classifiers(cls, classifiers=None):
        if classifiers is None:
            classifiers = cls.classifiers if cls.is_classified else ()
        return frozenset(t for t in classifiers if issubclass(t, SubtypeVar))

    @classmethod
    def non_wildcard_classifiers(cls, classifiers=None):
        if classifiers is None:
            classifiers = cls.classifiers if cls.is_classified else ()
        return frozenset(q for q in classifiers if not issubclass(q, SubtypeVar))

    @classmethod
    def __class_getitem__(cls, classifiers):
        """Set the content types for a newly created dynamically type"""
        if isinstance(classifiers, ty.Iterable):
            classifiers = tuple(classifiers)
        else:
            classifiers = (classifiers,)
        if cls.allowed_classifiers:
            not_allowed = [
                q
                for q in classifiers
                if not any(issubclass(q, t) for t in cls.allowed_classifiers)
            ]
            if not_allowed:
                raise FileFormatsError(
                    f"Invalid content types provided to {cls} (must be subclasses of "
                    f"{cls.allowed_classifiers}): {not_allowed}"
                )
        # Sort content types if order isn't important
        if cls.multiple_classifiers:
            if not cls.ordered_classifiers:
                # TODO: should check to see if classifiers are subclasses of each other
                repetitions = [t for t, c in Counter(classifiers).items() if c > 1]
                if repetitions:
                    raise FileFormatsError(
                        f"Cannot have more than one occurrence of a qualifier "
                        f"({repetitions}) for {cls} class when "
                        f"{cls.__name__}.ordered_classifiers is false"
                    )
                classifiers = frozenset(classifiers)
        else:
            if len(classifiers) > 1:
                raise FileFormatsError(
                    f"Multiple classifiers not permitted for {cls} types, provided: ({classifiers})"
                )
        # Make sure that the "classified" dictionary is present in this class not super
        # classes
        if "_classified_subtypes" not in cls.__dict__:
            cls._classified_subtypes = {}
        try:
            # Load previously created type so we can do ``assert MyType[Integer] is MyType[Integer]``
            classified = cls._classified_subtypes[classifiers]
        except KeyError:
            if not hasattr(cls, "classifiers_attr_name"):
                raise FileFormatsError(
                    f"{cls} needs to define the 'classifiers_attr_name' class attribute "
                    "with the name of the (different) class attribute to hold the "
                    "classified types"
                )
            if cls.classifiers_attr_name is None:
                raise FileFormatsError(
                    f"Inherited classifiers have been disabled in {cls} (by setting "
                    f'"classifiers_attr_name)" to None)'
                )
            try:
                classifiers_attr = getattr(cls, cls.classifiers_attr_name)
            except AttributeError:
                raise FileFormatsError(
                    f"Default value for classifiers attribute "
                    f"'{cls.classifiers_attr_name}' needs to be set in {cls}"
                )
            else:
                if classifiers_attr:
                    raise FileFormatsError(
                        f"Default value for classifiers attribute "
                        f"'{cls.classifiers_attr_name}' needs to be set in {cls}"
                    )
            class_attrs = {
                "unclassified": cls,
                "classifiers": classifiers,
            }
            class_attrs[cls.classifiers_attr_name] = (
                classifiers if cls.multiple_classifiers else classifiers[0]
            )
            qualifier_names = [t.__name__ for t in classifiers]
            if not cls.ordered_classifiers:
                qualifier_names.sort()
            classified = type(
                f"{'_'.join(qualifier_names)}__{cls.__name__}",
                (cls,),
                class_attrs,
            )
            cls._classified_subtypes[classifiers] = classified
        return classified

    @classmethod
    def get_converter_tuples(
        cls, source_format: type
    ) -> ty.List[ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]]]:
        """Search the registered converters to find an appropriate task and associated
        key-word args to perform the conversion between source and target formats

        Parameters
        ----------
        source_format : type(FileSet)
            the source format to convert from
        """
        # Try to see if a converter has been defined to the exact type
        available_converters = super().get_converter_tuples(
            source_format
        )  # pylint: disable=no-member
        # Failing that, see if there is a generic conversion between the container type
        # the source format (or subclass of) defined with matching wildcards in the source
        # and target formats
        if not available_converters and cls.is_classified:
            converters_dict = FileSet.get_converters_dict(
                cls.unclassified
            )  # pylint: disable=no-member
            for template_source_format, converter in converters_dict.items():
                if len(converter) == 3:  # was defined with wildcard classifiers
                    converter, conv_kwargs, template_classifiers = converter
                    # Attempt conversion from generic type to template match
                    if issubclass(template_source_format, SubtypeVar):
                        assert tuple(
                            cls.wildcard_classifiers(template_classifiers)
                        ) == (template_source_format,)
                        non_wildcards = cls.non_wildcard_classifiers(
                            template_classifiers
                        )
                        to_match = tuple(set(cls.classifiers).difference(non_wildcards))
                        if len(to_match) > 1:
                            wildcard_match = False
                        else:
                            wildcard_match = issubclass(source_format, to_match[0])
                    # Attempt template to template conversion match
                    elif getattr(source_format, "is_classified", False) and issubclass(
                        source_format.unclassified, template_source_format.unclassified
                    ):
                        assert cls.wildcard_classifiers(
                            template_classifiers
                        ) == cls.wildcard_classifiers(
                            template_source_format.classifiers
                        )
                        if cls.ordered_classifiers:
                            if len(cls.classifiers) != len(template_classifiers) or len(
                                source_format.classifiers
                            ) != len(template_source_format.classifiers):
                                wildcard_match = False
                            else:
                                wildcard_map = {}
                                for actual, template in zip(
                                    source_format.classifiers,
                                    template_source_format.classifiers,
                                ):
                                    if issubclass(template, SubtypeVar):
                                        wildcard_map[template] = actual
                                wildcard_match = True
                                for actual, template in zip(
                                    cls.classifiers, template_classifiers
                                ):
                                    if issubclass(template, SubtypeVar):
                                        try:
                                            reference = wildcard_map[template]
                                        except KeyError:
                                            wildcard_match = False
                                            break
                                        else:
                                            if not issubclass(actual, reference):
                                                wildcard_match = False
                                                break
                                    elif not issubclass(actual, template):
                                        wildcard_match = False
                                        break
                        else:
                            non_wildcards = cls.non_wildcard_classifiers(
                                template_classifiers
                            )
                            src_non_wildcards = cls.non_wildcard_classifiers(
                                template_source_format.classifiers
                            )
                            if not non_wildcards.issubset(
                                set(cls.classifiers)
                            ) or not src_non_wildcards.issubset(
                                set(source_format.classifiers)
                            ):
                                wildcard_match = False
                            else:
                                to_match = set(cls.classifiers).difference(
                                    non_wildcards
                                )
                                from_types = set(source_format.classifiers).difference(
                                    src_non_wildcards
                                )
                                wildcard_match = to_match.issubset(from_types)
                    else:
                        wildcard_match = False
                    if wildcard_match:
                        available_converters.append((converter, conv_kwargs))
        return available_converters

    @classmethod
    def __subclasshook__(cls, subclass: type) -> bool:
        """Overload the behaviour of 'issubclass' so that classified classes are considered
        to be subclasses of each other if they contain a super-set of classifiers"""
        if type.__subclasscheck__(cls, subclass):
            return True
        # Check to see whether the unclassified types are equivalent
        if (
            not cls.is_classified
            or not getattr(subclass, "is_classified", False)
            or not issubclass(subclass.unclassified, cls.unclassified)
        ):
            return False
        if cls.ordered_classifiers:
            assert subclass.ordered_classifiers
            if len(subclass.classifiers) != len(cls.classifiers):
                is_subclass = False
            else:
                is_subclass = all(
                    issubclass(q, s)
                    for q, s in zip(subclass.classifiers, cls.classifiers)
                )
        else:
            assert not subclass.ordered_classifiers
            if subclass.classifiers.issuperset(cls.classifiers):
                is_subclass = True
            else:
                # Check for sub-classes of classifiers
                is_subclass = all(
                    any(issubclass(q, s) for q in subclass.classifiers)
                    for s in cls.classifiers
                )
        return is_subclass

    @classmethod
    def register_converter(
        cls,
        source_format: type,
        converter_tuple: ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]],
    ):
        """Registers a converter task within a class attribute. Called by the @fileformats.mark.converter
        decorator.

        Parameters
        ----------
        source_format : type
            the source format to register a converter from
        task_spec : ty.Callable
            a callable that resolves to a Pydra task
        converter_kwargs : dict
            additional keyword arguments to be passed to the task spec at initialisation
            time

        Raises
        ------
        FormatConversionError
            if there is already a converter registered between the two types
        """
        # Ensure "converters" dict is defined in the target class and not in a superclass
        if cls.wildcard_classifiers():
            if issubclass(source_format, SubtypeVar):
                if len(cls.wildcard_classifiers()) > 1:
                    raise FileFormatsError(
                        "Can only have one wildcard qualifier when registering a converter "
                        f"to {cls} from a generic type, found {cls.wildcard_classifiers()}"
                    )
            elif not source_format.is_classified:
                raise FileFormatsError(
                    "Can only use wildcard classifiers when registering a converter "
                    f"from a generic type or similarly classified type, not {source_format}"
                )
            else:
                if cls.wildcard_classifiers() != source_format.wildcard_classifiers():
                    raise FileFormatsError(
                        f"Mismatching wildcards between source format, {source_format} "
                        f"({list(source_format.wildcard_classifiers())}), and target "
                        f"{cls} ({cls.wildcard_classifiers()})"
                    )
                prev_registered = [
                    f
                    for f in cls.converters
                    if (
                        issubclass(source_format.unclassified, f.unclassified)
                        and f.non_wildcard_classifiers()
                        == source_format.non_wildcard_classifiers()
                    )
                ]
                assert len(prev_registered) <= 1
                prev = prev_registered[0] if prev_registered else None
                if prev:
                    raise FileFormatsError(
                        f"There is already a converter registered from {prev.unclassified} "
                        f"to {cls.unclassified} with non-wildcard classifiers "
                        f"{list(prev.non_wildcard_classifiers())}: "
                        + describe_task(cls.converters[prev][0])
                    )
            converters_dict = cls.unclassified.get_converters_dict()
            converters_dict[source_format] = converter_tuple + (cls.classifiers,)
        else:
            super().register_converter(source_format, converter_tuple)

    @classproperty
    def namespace(cls):  # pylint: disable=no-self-argument
        """The "namespace" the format belongs to under the "fileformats" umbrella
        namespace"""
        if cls.is_classified:
            namespaces = set(t.namespace for t in cls.classifiers)
            if not cls.generically_qualifies:
                namespaces.add(cls.unclassified.namespace)
            if len(namespaces) == 1:
                return next(iter(namespaces))
            else:
                msg = (
                    "Cannot create reversible MIME type for because did not find a "
                    f"common namespace between all classifiers {list(cls.classifiers)}"
                )
                if cls.generically_qualifies:
                    msg += (
                        f" and (non genericly classified) base class {cls.unclassified}"
                    )
                raise FormatRecognitionError(msg + f", found:\n{list(namespaces)}")
        else:
            namespace = super().namespace
        return namespace

    @property
    def _type_name(self):
        """Name of type including classifiers to be used in __repr__"""
        if self.is_classified:
            unclassified = self.unclassified.__name__
        else:
            unclassified = type(self).__name__
        return (
            unclassified + "[" + ", ".join(t._type_name for t in self.classifiers) + "]"
        )



import typing as ty
from pathlib import Path
from fileformats.core.fileset import FileSet, MockMixin
from fileformats.testing import Foo


@FileSet.generate_sample_data.register
def generate_foo_data(file: Foo, dest_dir: Path) -> ty.List[Path]:
    foo_fspath = dest_dir / "test.foo"
    foo_fspath.write_text("foo")
    return [foo_fspath]


def test_sample():
    test_inst = Foo.sample()
    assert test_inst.fspath.exists()
    assert test_inst.fspath.name == "test.foo"


def test_mock():
    mock = Foo.mock()
    assert mock.fspath == Path("/mock/foo.foo")
    assert not mock.fspath.exists()
    assert isinstance(mock, MockMixin)



import typing as ty
import attrs
from fileformats.core import DataType
from fileformats.generic import File
from fileformats.field import Singular
from fileformats.core.mixin import WithClassifiers


class FileClassifier(DataType):
    pass


class A(FileClassifier):
    pass


class B(FileClassifier):
    pass


class C(FileClassifier):
    pass


class D(FileClassifier):
    pass


class E(C):
    pass


class F(WithClassifiers, File):
    classifiers_attr_name = "content_types"
    content_types = ()
    ext = ".f"


class G(F):
    ext = ".g"


class H(WithClassifiers, File):
    classifiers_attr_name = "content_types"
    content_types = ()
    ext = ".h"

    allowed_classifiers = (A, B, C)


class J(H):
    ext = ".j"


class K(WithClassifiers, File):

    ext = ".k"
    classifiers_attr_name = "new_classifiers_attr"
    new_classifiers_attr = ()
    ordered_classifiers = True


class L(WithClassifiers, File):

    ext = ".l"
    classifiers_attr_name = "new_classifiers_attr"
    new_classifiers_attr = ()
    ordered_classifiers = True


class M(WithClassifiers, File):
    classifiers_attr_name = "content_types"
    content_types = None  # Should be None not ()
    ext = ".m"
    multiple_classifiers = False


class N(WithClassifiers, File):
    classifiers_attr_name = "content_types"
    content_types = ()
    ext = ".n"


@attrs.define
class TestField(Singular):

    value: ty.Any


class P(WithClassifiers, File):
    ext = ".p"
    classifiers_attr_name = "content_types"
    content_types = ()


class Q(WithClassifiers, File):
    ext = ".z"
    classifiers_attr_name = "new_classifiers_attr"
    # MISSING default value for "new_classifiers_attr"


class R(WithClassifiers, File):

    ext = ".r"
    classifiers_attr_name = "new_classifiers_attr"
    new_classifiers_attr = ()
    ordered_classifiers = True



from pathlib import Path
import os.path
import random
import shutil
import typing as ty
import time
import pytest
from fileformats.core import FileSet, mark
from fileformats.generic import File, Directory, FsObject
from fileformats.core.mixin import WithSeparateHeader
from fileformats.core.utils import (
    find_matching,
    to_mime,
    from_mime,
)
from fileformats.core.exceptions import FileFormatsError
from fileformats.testing import Foo, Bar
import fileformats.text
from conftest import write_test_file


class Mario(File):
    ext = ".mario"


class Luigi(WithSeparateHeader, File):
    ext = ".luigi"
    header_type = Mario


class Bowser(Directory):
    content_types = (Mario,)


@pytest.fixture
def bowser_dir(work_dir):
    bowser_dir = work_dir / "bowser"
    bowser_dir.mkdir()
    for i in range(5):
        bar_fspath = bowser_dir / f"{i}.mario"
        write_test_file(bar_fspath)
    return Bowser(bowser_dir)


@pytest.fixture
def luigi_file(work_dir):
    luigi_fspath = work_dir / "x.luigi"
    write_test_file(luigi_fspath)
    bar_fspath = work_dir / "y.mario"
    write_test_file(bar_fspath)
    return Luigi([luigi_fspath, bar_fspath])


@pytest.fixture(params=["luigi", "bowser"])
def fsobject(luigi_file, bowser_dir, request):
    if request.param == "luigi":
        return luigi_file
    elif request.param == "bowser":
        return bowser_dir
    else:
        assert False


@pytest.fixture
def dest_dir(work_dir):
    dest_dir = work_dir / "new-dir"
    dest_dir.mkdir()
    return dest_dir


def test_copy(fsobject: FsObject, dest_dir: Path):
    cpy = fsobject.copy(dest_dir)
    assert all(p.parent == dest_dir for p in cpy.fspaths)
    assert set(p.name for p in cpy.fspaths) == set(p.name for p in fsobject.fspaths)
    assert cpy.hash() == fsobject.hash()


def test_copy_symlink(fsobject: FsObject, dest_dir: Path):
    cpy = fsobject.copy(dest_dir, mode=File.CopyMode.symlink)
    assert all(p.parent == dest_dir for p in cpy.fspaths)
    assert set(p.name for p in cpy.fspaths) == set(p.name for p in fsobject.fspaths)
    assert all(p.is_symlink() for p in cpy.fspaths)
    assert cpy.hash() == fsobject.hash()


def test_copy_symlink_with_fallback(fsobject: FsObject, dest_dir: Path):
    "Simulate source object on CIFS share, should be copied not symlinked"
    cpy = fsobject.copy(
        dest_dir,
        mode=File.CopyMode.link_or_copy,
        supported_modes=File.CopyMode.hardlink_or_copy,
    )
    assert all(p.parent == dest_dir for p in cpy.fspaths)
    assert set(p.name for p in cpy.fspaths) == set(p.name for p in fsobject.fspaths)
    assert not any(p.is_symlink() for p in cpy.fspaths)
    assert cpy.hash() == fsobject.hash()


def test_copy_symlink_with_unrequired_fallback(fsobject: FsObject, dest_dir: Path):
    "Simulate source object not on CIFS share, should be symlinked not copied"
    cpy = fsobject.copy(
        dest_dir, mode=File.CopyMode.link_or_copy, supported_modes=File.CopyMode.link
    )
    assert all(p.parent == dest_dir for p in cpy.fspaths)
    assert set(p.name for p in cpy.fspaths) == set(p.name for p in fsobject.fspaths)
    assert all(p.is_symlink() for p in cpy.fspaths)
    assert cpy.hash() == fsobject.hash()


def test_copy_hardlink(fsobject: FsObject, dest_dir: Path):
    cpy = fsobject.copy(dest_dir, mode=File.CopyMode.symlink)
    assert all(p.parent == dest_dir for p in cpy.fspaths)
    assert set(p.name for p in cpy.fspaths) == set(p.name for p in fsobject.fspaths)
    assert all(
        os.path.samefile(c, o)
        for c, o in zip(sorted(cpy.fspaths), sorted(fsobject.fspaths))
        if o.is_file()
    )
    assert cpy.hash() == fsobject.hash()


def test_copy_collation_same_name(work_dir: Path, dest_dir: Path):
    a = work_dir / "a" / "file.txt"
    b = work_dir / "b" / "file.txt"
    c = work_dir / "c" / "d" / "file.txt"
    fspaths = (a, b, c)

    for x in fspaths:
        x.parent.mkdir(parents=True)
        Path.touch(x)

    fileset = FileSet(fspaths)

    with pytest.raises(
        FileFormatsError,
        match="as there are duplicate filenames",
    ):
        fileset.copy(dest_dir=dest_dir, collation="siblings")
    cpy = fileset.copy(dest_dir=dest_dir, collation="any")
    assert sorted(cpy.relative_fspaths) == sorted(fileset.relative_fspaths)
    assert cpy.parent == dest_dir


def test_copy_collation_same_ext(work_dir: Path, dest_dir: Path):
    a = work_dir / "a" / "filea.txt"
    b = work_dir / "b" / "fileb.txt"
    c = work_dir / "c" / "d" / "filec.txt"
    fspaths = (a, b, c)

    for x in fspaths:
        x.parent.mkdir(parents=True)
        Path.touch(x)

    fileset = FileSet(fspaths)

    with pytest.raises(
        FileFormatsError,
        match="as there are duplicate extensions",
    ):
        fileset.copy(dest_dir=dest_dir, collation="adjacent")
    cpy = fileset.copy(dest_dir=dest_dir, collation="siblings")
    assert sorted(p.name for p in cpy.fspaths) == sorted(
        p.name for p in fileset.fspaths
    )
    assert all(p.parent == dest_dir for p in cpy.fspaths)


def test_copy_collation_diff_ext(work_dir: Path, dest_dir: Path):
    a = work_dir / "a" / "filea.x"
    b = work_dir / "b" / "fileb.y"
    c = work_dir / "c" / "d" / "filec.z"
    fspaths = (a, b, c)

    for x in fspaths:
        x.parent.mkdir(parents=True)
        Path.touch(x)

    fileset = FileSet(fspaths)

    cpy = fileset.copy(dest_dir=dest_dir, collation="adjacent")
    assert all(p.parent == dest_dir for p in cpy.fspaths)
    assert all(p.stem == "filea" for p in cpy.fspaths)


def test_copy_collation_stem(work_dir: Path, dest_dir: Path):
    a = work_dir / "a" / "filea.x"
    b = work_dir / "b" / "fileb.y"
    c = work_dir / "c" / "d" / "filec.z"
    fspaths = (a, b, c)

    for x in fspaths:
        x.parent.mkdir(parents=True)
        Path.touch(x)

    fileset = FileSet(fspaths)

    cpy = fileset.copy(dest_dir=dest_dir, collation="adjacent", new_stem="newfile")
    assert all(p.parent == dest_dir for p in cpy.fspaths)
    assert all(p.stem == "newfile" for p in cpy.fspaths)


def test_copy_collation_with_stem_not_adjacent(luigi_file: Luigi, dest_dir):
    with pytest.raises(
        FileFormatsError, match="when collation is not set to 'adjacent'"
    ):
        luigi_file.copy(dest_dir, collation="any", new_stem="new")


def test_copy_collation_leave_diff_dir(work_dir: Path, dest_dir: Path):
    a = work_dir / "a" / "file.x"
    b = work_dir / "b" / "file.y"
    c = work_dir / "c" / "file.z"
    fspaths = (a, b, c)

    for x in fspaths:
        x.parent.mkdir(parents=True)
        Path.touch(x)

    fileset = FileSet(fspaths)

    with pytest.raises(
        FileFormatsError,
        match="using leave mode as it is not supported by the any given the collation specification",
    ):
        fileset.copy(dest_dir=dest_dir, mode="leave", collation="siblings")

    with pytest.raises(
        FileFormatsError,
        match="using leave mode as it is not supported by the any given the collation specification",
    ):
        fileset.copy(dest_dir=dest_dir, mode="leave", collation="adjacent")


def test_copy_ext(work_dir):
    assert (
        File.copy_ext(
            work_dir / "x.luigi.mario",
            work_dir / "y",
            decomposition_mode=FileSet.ExtensionDecomposition.none,
        )
        == work_dir / "y"
    )
    assert (
        File.copy_ext(
            work_dir / "x.luigi.mario",
            work_dir / "y",
            decomposition_mode=FileSet.ExtensionDecomposition.single,
        )
        == work_dir / "y.mario"
    )
    assert (
        File.copy_ext(
            work_dir / "x.luigi.mario",
            work_dir / "y",
            decomposition_mode=FileSet.ExtensionDecomposition.multiple,
        )
        == work_dir / "y.luigi.mario"
    )
    assert (
        Mario.copy_ext(work_dir / "x.luigi.mario", work_dir / "y")
        == work_dir / "y.mario"
    )


def test_decompose_fspaths(work_dir):
    class LuigiMario(File):
        ext = ".luigi.mario"

    class DoubleMario(FileSet):
        @mark.required
        @property
        def bar(self):
            return Mario(self.select_by_ext(Mario))

        @mark.required
        @property
        def luigi_bar(self):
            return LuigiMario(self.select_by_ext(LuigiMario))

    fspath = work_dir / "file.luigi.mario"
    Path.touch(fspath)
    double_bar = DoubleMario(fspath)

    with pytest.warns(match="whereas it is also interpreted as a"):
        decomposed = double_bar.decomposed_fspaths()

    assert decomposed == [(work_dir, "file.luigi", ".mario")]


def test_format_detection(work_dir):
    text_file = work_dir / "text.txt"

    with open(text_file, "w") as f:
        f.write("sample text")

    detected = find_matching(text_file, standard_only=True)
    assert sorted(detected, key=lambda f: f.mime_like) == [
        fileformats.text.Prs_Fallenstein_Rst,
        fileformats.text.Prs_Prop_Logic,
        fileformats.text.TextFile,
    ]


def test_hash(tmp_path: Path):
    file_1 = tmp_path / "file_1.txt"
    file_2 = tmp_path / "file_2.txt"
    file_1.write_text("hello")
    file_2.write_text("hello")

    assert File(file_1).hash() == File(file_2).hash()


def test_hash_function_files_mismatch(tmp_path: Path):
    file_1 = tmp_path / "file_1.txt"
    file_2 = tmp_path / "file_2.txt"
    file_1.write_text("hello")
    file_2.write_text("hi")

    assert File(file_1).hash() != File(file_2).hash()


def test_hash_dir(tmp_path: Path):
    dir1 = tmp_path / "luigi"
    dir2 = tmp_path / "bar"
    for d in (dir1, dir2):
        d.mkdir()
        for i in range(3):
            f = d / f"{i}.txt"
            f.write_text(str(i))

    assert Directory(dir1).hash() == Directory(dir2).hash()


def test_hash_nested_dir(tmp_path: Path):
    dpath = tmp_path / "dir"
    dpath.mkdir()
    hidden = dpath / ".hidden"
    nested = dpath / "nested"
    hidden.mkdir()
    nested.mkdir()
    file_1 = dpath / "file_1.txt"
    file_2 = hidden / "file_2.txt"
    file_3 = nested / ".file_3.txt"
    file_4 = nested / "file_4.txt"

    for fx in [file_1, file_2, file_3, file_4]:
        fx.write_text(str(random.randint(0, 1000)))

    nested_dir = Directory(dpath)

    orig_hash = nested_dir.hash()

    nohidden_hash = nested_dir.hash(ignore_hidden_dirs=True, ignore_hidden_files=True)
    nohiddendirs_hash = nested_dir.hash(ignore_hidden_dirs=True)
    nohiddenfiles_hash = nested_dir.hash(ignore_hidden_files=True)

    assert orig_hash != nohidden_hash
    assert orig_hash != nohiddendirs_hash
    assert orig_hash != nohiddenfiles_hash

    os.remove(file_3)
    assert nested_dir.hash() == nohiddenfiles_hash
    shutil.rmtree(hidden)
    assert nested_dir.hash() == nohidden_hash


def test_hash_mtime(tmp_path: Path):
    file_1 = tmp_path / "file_1.txt"
    file_2 = tmp_path / "file_2.txt"
    file_1.write_text("hello")
    file_2.write_text("hello")

    orig_hash = File(file_1).hash(mtime=True)

    assert File(file_1).hash(mtime=True) == orig_hash
    assert File(file_2).hash(mtime=True) != orig_hash

    time.sleep(1)
    Path.touch(file_1)
    assert File(file_1).hash(mtime=True) != orig_hash


def test_hash_files(fsobject: FsObject, work_dir: Path, dest_dir: Path):
    file_hashes = fsobject.hash_files(relative_to=work_dir)
    assert sorted(Path(p) for p in file_hashes) == sorted(
        p.relative_to(work_dir) for p in fsobject.all_file_paths
    )
    cpy = fsobject.copy(dest_dir)
    assert cpy.hash_files() == fsobject.hash_files()


def test_to_from_mime_roundtrip():
    mime_str = to_mime(Foo)
    assert isinstance(mime_str, str)
    assert from_mime(mime_str) == Foo


def test_to_from_list_mime_roundtrip():
    mime_str = to_mime(ty.List[Foo])
    assert isinstance(mime_str, str)
    assert from_mime(mime_str) == ty.List[Foo]


def test_to_from_union_mime_roundtrip():
    mime_str = to_mime(ty.Union[Foo, Bar])
    assert isinstance(mime_str, str)
    assert from_mime(mime_str) == ty.Union[Foo, Bar]


def test_to_from_list_union_mime_roundtrip():
    mime_str = to_mime(ty.List[ty.Union[Foo, Bar]])
    assert isinstance(mime_str, str)
    assert from_mime(mime_str) == ty.List[ty.Union[Foo, Bar]]


def test_official_mime_fail():
    with pytest.raises(TypeError, match="as it is not a proper file-type"):
        to_mime(ty.List[Foo], official=True)



from .medical import Dicom
from ..text import Javascript
from .unclassified import (
    _1dInterleavedParityfec,
    _3gpdashQoeReport__Xml,
    _3gpphal__Json,
    _3gpphalforms__Json,
    _3gppIms__Xml,
    A2l,
    Ace__Cbor,
    Ace__Json,
    Activemessage,
    Activity__Json,
    Aif__Cbor,
    Aif__Json,
    AltoCdni__Json,
    AltoCdnifilter__Json,
    AltoCostmap__Json,
    AltoCostmapfilter__Json,
    AltoDirectory__Json,
    AltoEndpointprop__Json,
    AltoEndpointpropparams__Json,
    AltoEndpointcost__Json,
    AltoEndpointcostparams__Json,
    AltoError__Json,
    AltoNetworkmapfilter__Json,
    AltoNetworkmap__Json,
    AltoPropmap__Json,
    AltoPropmapparams__Json,
    AltoUpdatestreamcontrol__Json,
    AltoUpdatestreamparams__Json,
    Aml,
    AndrewInset,
    Applefile,
    At__Jwt,
    Atf,
    Atfx,
    Atom__Xml,
    Atomcat__Xml,
    Atomdeleted__Xml,
    Atomicmail,
    Atomsvc__Xml,
    AtscDwd__Xml,
    AtscDynamicEventMessage,
    AtscHeld__Xml,
    AtscRdt__Json,
    AtscRsat__Xml,
    Atxml,
    AuthPolicy__Xml,
    AutomationmlAml__Xml,
    AutomationmlAmlx__Zip,
    BacnetXdd__Zip,
    BatchSmtp,
    Beep__Xml,
    Calendar__Json,
    Calendar__Xml,
    CallCompletion,
    Cals_1840,
    Captive__Json,
    Cbor,
    CborSeq,
    Cccex,
    Ccmp__Xml,
    Ccxml__Xml,
    Cda__Xml,
    Cdfx__Xml,
    CdmiCapability,
    CdmiContainer,
    CdmiDomain,
    CdmiObject,
    CdmiQueue,
    Cdni,
    Cea,
    Cea_2018__Xml,
    Cellml__Xml,
    Cfw,
    City__Json,
    Clr,
    ClueInfo__Xml,
    Clue__Xml,
    Cms,
    Cnrp__Xml,
    CoapGroup__Json,
    CoapPayload,
    Commonground,
    ConciseProblemDetails__Cbor,
    ConferenceInfo__Xml,
    Cpl__Xml,
    Cose,
    CoseKey,
    CoseKeySet,
    CoseX509,
    Csrattrs,
    Csta__Xml,
    Cstadata__Xml,
    Csvm__Json,
    Cwl,
    Cwl__Json,
    Cwt,
    Cybercash,
    Dash__Xml,
    DashPatch__Xml,
    Dashdelta,
    Davmount__Xml,
    DcaRft,
    Dcd,
    DecDx,
    DialogInfo__Xml,
    Dicom__Json,
    Dicom__Xml,
    Dii,
    Dit,
    Dns,
    Dns__Json,
    DnsMessage,
    Dots__Cbor,
    Dpop__Jwt,
    Dskpp__Xml,
    Dssc__Der,
    Dssc__Xml,
    Dvcs,
    EdiConsent,
    Edifact,
    EdiX12,
    Efi,
    Elm__Json,
    Elm__Xml,
    Emergencycalldata_Cap__Xml,
    Emergencycalldata_Comment__Xml,
    Emergencycalldata_Control__Xml,
    Emergencycalldata_Deviceinfo__Xml,
    Emergencycalldata_Ecall_Msd,
    Emergencycalldata_Legacyesn__Json,
    Emergencycalldata_Providerinfo__Xml,
    Emergencycalldata_Serviceinfo__Xml,
    Emergencycalldata_Subscriberinfo__Xml,
    Emergencycalldata_Veds__Xml,
    Emma__Xml,
    Emotionml__Xml,
    Encaprtp,
    Epp__Xml,
    Epub__Zip,
    Eshop,
    Exi,
    ExpectCtReport__Json,
    Express,
    Fastinfoset,
    Fastsoap,
    Fdf,
    Fdt__Xml,
    Fhir__Json,
    Fhir__Xml,
    Fits,
    Flexfec,
    FrameworkAttributes__Xml,
    Geo__Json,
    Geo__JsonSeq,
    Geopackage__Sqlite3,
    Geoxacml__Xml,
    GltfBuffer,
    Gml__Xml,
    H224,
    Held__Xml,
    Hl7v2__Xml,
    Http,
    Hyperstudio,
    IbeKeyRequest__Xml,
    IbePkgReply__Xml,
    IbePpData,
    Iges,
    ImIscomposing__Xml,
    Index,
    Index_Cmd,
    Index_Obj,
    Index_Response,
    Index_Vnd,
    Inkml__Xml,
    Iotp,
    Ipfix,
    Ipp,
    Isup,
    Its__Xml,
    JavaArchive,
    Jf2feed__Json,
    Jose,
    Jose__Json,
    Jrd__Json,
    Jscalendar__Json,
    JsonPatch__Json,
    JsonSeq,
    Jwk__Json,
    JwkSet__Json,
    Jwt,
    KpmlRequest__Xml,
    KpmlResponse__Xml,
    Ld__Json,
    Lgr__Xml,
    LinkFormat,
    Linkset,
    Linkset__Json,
    LoadControl__Xml,
    Logout__Jwt,
    Lost__Xml,
    Lostsync__Xml,
    Lpf__Zip,
    Lxf,
    MacBinhex40,
    Macwriteii,
    Mads__Xml,
    Manifest__Json,
    Marc,
    Marcxml__Xml,
    Mathematica,
    Mathml__Xml,
    MathmlContent__Xml,
    MathmlPresentation__Xml,
    MbmsAssociatedProcedureDescription__Xml,
    MbmsDeregister__Xml,
    MbmsEnvelope__Xml,
    MbmsMskResponse__Xml,
    MbmsMsk__Xml,
    MbmsProtectionDescription__Xml,
    MbmsReceptionReport__Xml,
    MbmsRegisterResponse__Xml,
    MbmsRegister__Xml,
    MbmsSchedule__Xml,
    MbmsUserServiceDescription__Xml,
    Mbox,
    MediaControl__Xml,
    MediaPolicyDataset__Xml,
    Mediaservercontrol__Xml,
    MergePatch__Json,
    Metalink4__Xml,
    Mets__Xml,
    Mf4,
    Mikey,
    Mipc,
    MissingBlocks__CborSeq,
    MmtAei__Xml,
    MmtUsd__Xml,
    Mods__Xml,
    MossKeys,
    MossSignature,
    MosskeyData,
    MosskeyRequest,
    Mp21,
    Mp4,
    Mpeg4Generic,
    Mpeg4Iod,
    Mpeg4IodXmt,
    MrbConsumer__Xml,
    MrbPublish__Xml,
    MscIvr__Xml,
    MscMixer__Xml,
    Mud__Json,
    MultipartCore,
    Mxf,
    NQuads,
    NTriples,
    Nasdata,
    NewsCheckgroups,
    NewsGroupinfo,
    NewsTransmission,
    Nlsml__Xml,
    Node,
    Nss,
    OauthAuthzReq__Jwt,
    ObliviousDnsMessage,
    OcspRequest,
    OcspResponse,
    OctetStream,
    Oda,
    Odm__Xml,
    Odx,
    OebpsPackage__Xml,
    Ogg,
    OhttpKeys,
    OpcNodeset__Xml,
    Oscore,
    Oxps,
    P21,
    P21__Zip,
    P2pOverlay__Xml,
    Parityfec,
    Passport,
    PatchOpsError__Xml,
    Pdx,
    PemCertificateChain,
    PgpEncrypted,
    PgpKeys,
    PgpSignature,
    PidfDiff__Xml,
    Pidf__Xml,
    Pkcs10,
    Pkcs7Mime,
    Pkcs7Signature,
    Pkcs8,
    Pkcs8Encrypted,
    Pkcs12,
    PkixAttrCert,
    PkixCert,
    PkixCrl,
    PkixPkipath,
    Pkixcmp,
    Pls__Xml,
    PocSettings__Xml,
    PpspTracker__Json,
    Problem__Json,
    Problem__Xml,
    Provenance__Xml,
    Prs_Alvestrand_TitraxSheet,
    Prs_Cww,
    Prs_Cyn,
    Prs_Hpub__Zip,
    Prs_ImpliedDocument__Xml,
    Prs_ImpliedExecutable,
    Prs_ImpliedStructure,
    Prs_Nprend,
    Prs_Plucker,
    Prs_RdfXmlCrypt,
    Prs_Xsf__Xml,
    Pskc__Xml,
    Pvd__Json,
    Rdf__Xml,
    RouteApd__Xml,
    RouteSTsid__Xml,
    RouteUsd__Xml,
    Qsig,
    Raptorfec,
    Rdap__Json,
    Reginfo__Xml,
    RelaxNgCompactSyntax,
    Reputon__Json,
    ResourceListsDiff__Xml,
    ResourceLists__Xml,
    Rfc__Xml,
    Riscos,
    Rlmi__Xml,
    RlsServices__Xml,
    RpkiChecklist,
    RpkiGhostbusters,
    RpkiManifest,
    RpkiPublication,
    RpkiRoa,
    RpkiUpdown,
    Rtf,
    Rtploopback,
    Rtx,
    Samlassertion__Xml,
    Samlmetadata__Xml,
    SarifExternalProperties__Json,
    Sarif__Json,
    Sbe,
    Sbml__Xml,
    Scaip__Xml,
    Scim__Json,
    ScvpCvRequest,
    ScvpCvResponse,
    ScvpVpRequest,
    ScvpVpResponse,
    Sdp,
    Secevent__Jwt,
    SenmlEtch__Cbor,
    SenmlEtch__Json,
    SenmlExi,
    Senml__Cbor,
    Senml__Json,
    Senml__Xml,
    SensmlExi,
    Sensml__Cbor,
    Sensml__Json,
    Sensml__Xml,
    SepExi,
    Sep__Xml,
    SessionInfo,
    SetPayment,
    SetPaymentInitiation,
    SetRegistration,
    SetRegistrationInitiation,
    Sgml,
    SgmlOpenCatalog,
    Shf__Xml,
    Sieve,
    SimpleFilter__Xml,
    SimpleMessageSummary,
    Simplesymbolcontainer,
    Sipc,
    Slate,
    Smil__Xml,
    Smpte336m,
    Soap__Fastinfoset,
    Soap__Xml,
    SparqlQuery,
    Spdx__Json,
    SparqlResults__Xml,
    SpiritsEvent__Xml,
    Sql,
    Srgs,
    Srgs__Xml,
    Sru__Xml,
    Ssml__Xml,
    Stix__Json,
    Swid__Cbor,
    Swid__Xml,
    TampApexUpdate,
    TampApexUpdateConfirm,
    TampCommunityUpdate,
    TampCommunityUpdateConfirm,
    TampError,
    TampSequenceAdjust,
    TampSequenceAdjustConfirm,
    TampStatusQuery,
    TampStatusResponse,
    TampUpdate,
    TampUpdateConfirm,
    Taxii__Json,
    Td__Json,
    Tei__Xml,
    TetraIsi,
    Thraud__Xml,
    TimestampQuery,
    TimestampReply,
    TimestampedData,
    Tlsrpt__Gzip,
    Tlsrpt__Json,
    Tm__Json,
    Tnauthlist,
    TokenIntrospection__Jwt,
    TrickleIceSdpfrag,
    Trig,
    Ttml__Xml,
    TveTrigger,
    Tzif,
    TzifLeap,
    Ulpfec,
    UrcGrpsheet__Xml,
    UrcRessheet__Xml,
    UrcTargetdesc__Xml,
    UrcUisocketdesc__Xml,
    Vcard__Json,
    Vcard__Xml,
    Vemmi,
    Voicexml__Xml,
    VoucherCms__Json,
    VqRtcpxr,
    Wasm,
    Watcherinfo__Xml,
    WebpushOptions__Json,
    WhoisppQuery,
    WhoisppResponse,
    Widget,
    Wita,
    Wordperfect5_1,
    Wsdl__Xml,
    Wspolicy__Xml,
    PkiMessage,
    WwwFormUrlencoded,
    X509CaCert,
    X509CaRaCert,
    X509NextCaCert,
    X400Bp,
    Xacml__Xml,
    XcapAtt__Xml,
    XcapCaps__Xml,
    XcapDiff__Xml,
    XcapEl__Xml,
    XcapError__Xml,
    XcapNs__Xml,
    XconConferenceInfoDiff__Xml,
    XconConferenceInfo__Xml,
    Xenc__Xml,
    Xfdf,
    Xhtml__Xml,
    Xliff__Xml,
    Xml,
    XmlDtd,
    XmlExternalParsedEntity,
    XmlPatch__Xml,
    Xmpp__Xml,
    Xop__Xml,
    Xslt__Xml,
    Xv__Xml,
    Yang,
    YangData__Cbor,
    YangData__Json,
    YangData__Xml,
    YangPatch__Json,
    YangPatch__Xml,
    Yin__Xml,
    Zlib,
    Zstd,
)



from fileformats.core import __version__
from warnings import warn
from fileformats.core import FileSet
from fileformats.core.mixin import WithClassifiers
from fileformats.generic import File
from fileformats.core.mixin import WithMagicNumber


class Archive(WithClassifiers, File):
    "Base class for compressed archives"

    classifiers_attr_name = "archived_type"
    archived_type = None
    multiple_classifiers = False
    allowed_classifiers = (FileSet,)
    generically_qualifies = True

    binary = True
    iana_mime = None


# Compressed formats
class Zip(WithMagicNumber, Archive):
    ext = ".zip"
    magic_number = "504B0304"
    iana_mime = "application/zip"


class Bzip(WithMagicNumber, Archive):
    ext = ".bzip"
    magic_number = "425a"
    iana_mime = "application/bzip"


class Gzip(WithMagicNumber, Archive):
    ext = ".gz"
    magic_number = "1F8B08"
    iana_mime = "application/gzip"


class Tar(WithMagicNumber, Archive):
    ext = ".tar"
    magic_number = "7573746172"
    magic_number_offset = 257
    iana_mime = "application/x-tar"


class TarGzip(WithMagicNumber, Archive):
    # FIXME: Should capture the relationship to Gzip and Tar somehow, but a bit
    # tricky to get it right
    ext = ".tar.gz"
    magic_number = "1F8B08"
    alternate_exts = (".tgz",)
    iana_mime = "application/x-tar+gzip"



from fileformats.generic import File
from fileformats.core.mixin import WithMagicNumber


class Dicom(WithMagicNumber, File):

    iana_mime = "application/dicom"
    magic_number = b"DICM"
    magic_number_offset = 128
    binary = True

    alternate_exts = (".dcm",)  # dcm is recommended not required



from .base import Image
from fileformats.serialization import InformalSchema, Xml


class VectorImage(Image):
    iana_mime = None


class Svg(InformalSchema):
    pass


class Svg__Xml(VectorImage, Xml):
    ext = ".svg"
    iana_mime = "image/svg+xml"
    schema = Svg  # Set Xml schema attribute
    classifiers_attr_name = None  # disable classifiers inherited from Xml



from fileformats.core.utils import (
    include_testing_package,
    subpackages,
)


def test_include_testing():
    assert "fileformats.testing" not in [
        p.__name__ for p in subpackages(exclude=("core", "testing"))
    ]
    assert "fileformats.testing" in [p.__name__ for p in subpackages(exclude=("core",))]


def test_include_testing_flag():
    assert "fileformats.testing" in [p.__name__ for p in subpackages()]
    include_testing_package(False)
    try:
        pkgs = list(subpackages())
    finally:
        include_testing_package(True)
    assert "fileformats.testing" not in [p.__name__ for p in pkgs]
    assert "fileformats.testing" in [p.__name__ for p in subpackages()]



from __future__ import annotations
import typing as ty
import attrs
from .utils import (
    classproperty,
)
from .exceptions import (
    FileFormatsError,
)
from .datatype import DataType


@attrs.define(repr=False)
class Field(DataType):
    value = attrs.field()

    type = None
    is_field = True
    primitive = None

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return f"{self._type_name}({str(self)})"

    @property
    def metadata(self):
        """Empty metadata dict for duck-typing with file-sets"""
        return {}

    @classproperty
    def all_fields(cls) -> list[ty.Type[Field]]:  # pylint: disable=no-self-argument
        """Iterate over all field formats in fileformats.* namespaces"""
        import fileformats.field  # noqa

        return [f for f in Field.subclasses() if f.primitive is not None]

    @classmethod
    def from_primitive(cls, dtype: type):
        try:
            datatype = next(iter(f for f in cls.all_fields if f.primitive is dtype))
        except StopIteration as e:
            field_types_str = ", ".join(t.__name__ for t in cls.all_fields)
            raise FileFormatsError(
                f"{dtype} doesn't not correspond to a valid fileformats field type "
                f"({field_types_str})"
            ) from e
        return datatype

    _all_fields = None



import shutil
import pytest
from imageio.core.fetching import get_remote_file
from fileformats.image import Png, Tiff
from fileformats.core.exceptions import FormatMismatchError


@pytest.fixture(scope="session")
def png_path():
    return get_remote_file("images/chelsea.png")


def test_tiff_endianness(png_path):
    png = Png(png_path)
    Tiff.convert(png)


def test_tiff_fail(png_path, work_dir):
    bad_tiff_path = work_dir / "bad_tiff.tiff"
    shutil.copyfile(png_path, bad_tiff_path)
    with pytest.raises(FormatMismatchError):
        Tiff(bad_tiff_path)



import pytest
from fileformats.core.exceptions import FormatMismatchError
from fileformats.field import Text, Integer, Decimal, Boolean, Array


def test_field_text():
    text = Text("a-field")
    assert str(text) == "a-field"


def test_field_integer():
    integer = Integer("10")
    assert str(integer) == "10"
    assert int(integer) == 10
    assert integer + 2 == 12
    assert integer * 2 == 20
    assert integer // 2 == 5
    assert integer / 4 == 2.5
    assert integer % 7 == 3
    assert integer**2 == 100
    assert -integer == -10
    assert abs(-integer) == 10


def test_field_integer_fail():
    with pytest.raises(FormatMismatchError):
        Integer("foo")
    with pytest.raises(FormatMismatchError):
        Integer("1.5")


def test_field_decimal():
    decimal = Decimal("20.2")
    assert str(decimal) == "20.2"
    assert float(decimal) == 20.2


def test_field_decimal_fail():
    with pytest.raises(FormatMismatchError):
        Decimal("foo")


def test_field_boolean():
    assert Boolean("True")
    assert Boolean("true")
    assert Boolean("yes")
    assert Boolean("1")
    assert Boolean("YES")
    assert not Boolean("False")
    assert not Boolean("false")
    assert not Boolean("no")
    assert not Boolean("0")
    assert not Boolean("NO")
    boolean = Boolean(True)
    assert boolean and False is False
    assert boolean or False is True
    assert boolean is not False
    assert ~boolean is False


def test_field_boolean_fail():
    with pytest.raises(FormatMismatchError):
        Boolean("foo")
    with pytest.raises(FormatMismatchError):
        Boolean("1.0")


def test_field_array():
    array = Array("[1,2,3,4,5]")
    assert str(array) == "[1,2,3,4,5]"
    array = Array("[1, 2, 3, 4, 5]")
    assert str(array) == "[1,2,3,4,5]"
    array = Array("1,2,3,4,5")
    assert str(array) == "[1,2,3,4,5]"
    assert list(array) == ["1", "2", "3", "4", "5"]
    array = Array[Integer]("1,2,3,4,5")
    assert str(array) == "[1,2,3,4,5]"
    assert list(array) == [1, 2, 3, 4, 5]


def test_field_array_fail():
    with pytest.raises(FormatMismatchError):
        Boolean("1,2,3]")
    with pytest.raises(FormatMismatchError):
        Boolean("[1,2,3")



from pathlib import Path
import pytest
from fileformats.generic import File
from fileformats.field import Integer, Boolean, Decimal, Array, Text
from fileformats.testing import Foo
from fileformats.core.exceptions import FileFormatsError, FormatMismatchError
from fileformats.core import mark
from conftest import write_test_file


def test_init_args(work_dir):
    fspath = work_dir / "test.txt"
    write_test_file(fspath)
    File([fspath])
    with pytest.raises(TypeError):
        File(fspath, fspath)


class TestFile(File):

    ext = ".tst"


def test_file_repr():
    assert repr(Foo.mock()) == "Foo('/mock/foo.foo')"


def test_field_repr():
    assert repr(Integer(1)) == "Integer(1)"
    assert repr(Decimal("1.1")) == "Decimal(1.1)"
    assert repr(Boolean(1)) == "Boolean(true)"
    assert repr(Text(1)) == 'Text("1")'
    assert repr(Array[Integer]([1, 2, 3])) == "Array[Integer]([1,2,3])"


def test_multiple_matches1(work_dir):
    fspath1 = work_dir / "test1.tst"
    write_test_file(fspath1)
    fspath2 = work_dir / "test2.tst"
    write_test_file(fspath2)
    with pytest.raises(FormatMismatchError):
        File([fspath1, fspath2])


def test_multiple_matches2(work_dir):
    fspath1 = work_dir / "test1.tst"
    write_test_file(fspath1)
    fspath2 = work_dir / "test2.tst"
    write_test_file(fspath2)
    with pytest.raises(FormatMismatchError):
        TestFile([fspath1, fspath2])


def test_missing_files(work_dir):
    fspath = work_dir / "test.tst"
    write_test_file(fspath)
    with pytest.raises(FileNotFoundError):
        TestFile([fspath, work_dir / "missing1.txt", work_dir / "missing2.txt"])


def test_python_hash_fileset(work_dir: Path):

    a_path = work_dir / "a.txt"
    a_path.write_text("a")
    a = File(a_path)
    a2 = File(a_path)
    b_path = work_dir / "b.txt"
    b_path.write_text("b")
    b = File(b_path)

    assert hash(a) == hash(a2)
    assert hash(b) != hash(a)


def test_python_hash_integer():

    a = Integer(1)
    a2 = Integer(1)
    b = Integer(2)

    assert hash(a) == hash(a2)
    assert hash(b) != hash(a)


def test_python_hash_decimal():

    a = Decimal(1)
    a2 = Decimal(1)
    b = Decimal(2)

    assert hash(a) == hash(a2)
    assert hash(b) != hash(a)


def test_python_hash_boolean():

    a = Boolean(1)
    a2 = Boolean(1)
    b = Boolean(0)

    assert hash(a) == hash(a2)
    assert hash(b) != hash(a)


def test_python_hash_text():

    a = Text("1")
    a2 = Text("1")
    b = Text("2")

    assert hash(a) == hash(a2)
    assert hash(b) != hash(a)


def test_python_hash_array():

    a = Array[Integer]([1, 2])
    a2 = Array[Integer]([1, 2])
    b = Array[Integer]([1, 2, 3])

    assert hash(a) == hash(a2)
    assert hash(b) != hash(a)


class ImageWithInlineHeader(File):

    ext = ".img"
    binary = True

    header_separator = b"---END HEADER---"

    def read_metadata(self):
        hdr = self.contents.split(self.header_separator)[0].decode("utf-8")
        return {k: int(v) for k, v in (ln.split(":") for ln in hdr.splitlines())}


@pytest.mark.xfail(reason="Disabled separate header metadata")
def test_header_overwrite(work_dir):

    fspath = work_dir / "image.img"
    hdr = {
        "x": 10,
        "y": 20,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    file = ImageWithInlineHeader(fspath, metadata={"x": "100"})
    with pytest.raises(FileFormatsError):
        file.metadata["y"]


class YFile(ImageWithInlineHeader):
    @mark.required
    @property
    def y(self):
        return self.metadata["y"]

    @mark.check
    def y_value(self):
        if self.y <= 10:
            raise FormatMismatchError(f"'y' property is not > 10 ({self.y})")


def test_required_check_op(work_dir):
    fspath = work_dir / "image.img"
    hdr = {
        "x": 10,
        "y": 20,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    assert YFile.matches(fspath)


def test_required_check_op_fail(work_dir):
    fspath = work_dir / "image.img"
    hdr = {
        "x": 10,
        "y": 10,
    }
    write_test_file(
        fspath,
        (
            "\n".join(f"{k}:{v}" for k, v in hdr.items()).encode("utf-8")
            + ImageWithInlineHeader.header_separator
            + b"SOMERANDOMBYTESDATA"
        ),
        binary=True,
    )
    assert not YFile.matches(fspath)



from __future__ import annotations
import os
from copy import copy
import struct
from enum import Enum, IntEnum
from warnings import warn
import tempfile
from collections import Counter
import typing as ty
import shutil
from operator import itemgetter
import itertools
import functools
from pathlib import Path
import hashlib
import logging
import attrs
from .utils import (
    classproperty,
    fspaths_converter,
    to_mime_format_name,
    STANDARD_NAMESPACES,
    describe_task,
    import_extras_module,
)
from .converter import SubtypeVar
from .exceptions import (
    FileFormatsError,
    FormatMismatchError,
    FormatConversionError,
    FileFormatsExtrasError,
    FileFormatsExtrasPkgUninstalledError,
    FileFormatsExtrasPkgNotCheckedError,
)
from .datatype import DataType
from . import mark

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self


FILE_CHUNK_LEN_DEFAULT = 8192


logger = logging.getLogger("fileformats")


@attrs.define(slots=False, repr=False)
class FileSet(DataType):
    """
    The base class for all format types within the fileformats package. A generic
    representation of a collection of files related to a single data resource. A
    file-set can be a single file or directory or a collection thereof, such as a
    primary file with a "side-car" header.

    Parameters
    ----------
    fspaths : set[Path]
        a set of file-system paths pointing to all the resources in the file-set
    checks : bool
        whether to run in-depth "checks" to verify the file format
    """

    fspaths: ty.FrozenSet[Path] = attrs.field(default=None, converter=fspaths_converter)

    # Explicitly set the Internet Assigned Numbers Authority (https://iana_mime.org) MIME
    # type to None for any base classes that should not correspond to a MIME or MIME-like
    # type.
    iana_mime = None
    ext = None
    alternate_exts = ()

    # Store converters registered by @converter decorator that convert to FileSet
    # NB: each class will have its own version of this dictionary
    converters = {}

    is_fileset = True

    def __hash__(self):
        return hash(self.fspaths)

    def __repr__(self):
        return f"{self._type_name}('" + "', '".join(str(p) for p in self.fspaths) + "')"

    def __attrs_post_init__(self):
        # Check required properties don't raise errors
        for prop_name in self.required_properties():
            getattr(self, prop_name)
        # Loop through all attributes and find methods marked by CHECK_ANNOTATION
        for check in self.checks():
            getattr(self, check)()

    @fspaths.validator
    def validate_fspaths(self, _, fspaths):
        if not fspaths:
            raise FileFormatsError(f"No file-system paths provided to {self}")
        missing = [p for p in fspaths if not p or not p.exists()]
        if missing:
            missing_str = "\n".join(str(p) for p in missing)
            msg = (
                f"The following file system paths provided to {type(self)} do not "
                f"exist:\n\n{missing_str}\n\n"
            )
            present_parents = set()
            for fspath in missing:
                if fspath:
                    if fspath.parent.exists():
                        present_parents.add(fspath.parent)
            for parent in present_parents:
                msg += (
                    f"\n\nFiles in the present parent directory '{str(parent)}' are:\n"
                )
                msg += "\n".join(str(p) for p in parent.iterdir())
            raise FileNotFoundError(msg)

    @property
    def metadata(self):
        """Lazily load metadata from `read_metadata` extra if implemented, returning an
        empty metadata array if not"""
        try:
            metadata = self._metadata
        except AttributeError:
            try:
                self._metadata = self.read_metadata()
            except FileFormatsExtrasPkgUninstalledError:
                raise
            except FileFormatsExtrasPkgNotCheckedError as e:
                logger.warning(str(e))
                metadata = {}
            except FileFormatsExtrasError:
                metadata = {}
            else:
                metadata = self._metadata
        return metadata

    @mark.extra
    def read_metadata(self) -> ty.Dict[str, ty.Any]:
        """Reads any metadata associated with the fileset and returns it as a dict"""
        raise NotImplementedError

    @property
    def parent(self) -> Path:
        "A common parent directory for all the top-level paths in the file-set"
        return Path(os.path.commonpath(p.parent for p in self.fspaths))

    @property
    def relative_fspaths(self) -> ty.Iterator[Path]:
        "Paths for all top-level paths in the file-set relative to the common parent directory"
        return (p.relative_to(self.parent) for p in self.fspaths)

    @classproperty
    def mime_type(cls):
        """Generates a MIME type (IANA) identifier from a format class. If an official
        IANA MIME type doesn't exist it will create one in the in the MIME style, e.g.

            fileformats.text.Plain to "text/plain"

            fileformats.image.TiffFx to "image/tiff-fx"

            fileformats.mynamespace.MyFormat to "application/x-my-format

        Returns
        -------
        str
            the MIME type corresponding to the class
        """
        if getattr(cls, "iana_mime", None) is not None:
            mime_type = cls.iana_mime
        else:
            format_name = to_mime_format_name(cls.__name__)
            mime_type = f"application/x-{format_name}"
        return mime_type

    @classproperty
    def strext(cls) -> str:
        """Return extension that is guaranteed to be a string (i.e. not None)"""
        return cls.ext if cls.ext is not None else ""

    @classproperty
    def unconstrained(cls) -> bool:
        """Whether the file-format is unconstrained by extension, magic number or another
        constraint"""
        return not list(cls.required_properties())

    @classproperty
    def possible_exts(cls):
        possible = [cls.ext]
        try:
            possible.extend(cls.alternate_exts)
        except AttributeError:
            pass
        return possible

    @classmethod
    def required_properties(cls):
        """Find all properties required to treat file-set as being in the format specified
        by the class

        Returns
        -------
        iter(str)
            an iterator over all properties names marked as "required"
        """
        fileset_props = dir(FileSet)
        for attr_name in dir(cls):
            if attr_name in fileset_props:
                continue
            klass_attr = getattr(cls, attr_name)
            if isinstance(klass_attr, property):
                try:
                    klass_attr.fget.__annotations__[cls.REQUIRED_ANNOTATION]
                except KeyError:
                    pass
                else:
                    yield attr_name

    def required_paths(self) -> ty.Set[Path]:
        """Returns all fspaths that are required for the format"""
        required = set()
        for prop_name in self.required_properties():
            prop = getattr(self, prop_name)
            paths = []
            if hasattr(prop, "required_paths"):
                paths = prop.required_paths()
            elif isinstance(prop, os.PathLike):
                paths = [Path(prop)]
            elif isinstance(prop, ty.Iterable):
                for p in prop:
                    if isinstance(p, os.PathLike):
                        paths.append(Path(p))
            for path in paths:
                if path in self.fspaths:
                    required.add(path)
        return required

    def nested_filesets(self) -> ty.List[FileSet]:
        """Returns all nested filesets that are required for the format

        Returns
        ------
        fileset : list[FileSet]
            a fileset that is nested within the broader fileset
        """
        nested = []
        for prop_name in sorted(self.required_properties()):
            prop = getattr(self, prop_name)
            if isinstance(prop, FileSet):
                nested.append(prop)
                nested.extend(prop.nested_filesets())
        return nested

    def trim_paths(self):
        """Trims paths in fspaths to only those that are "required" by the format class
        i.e. returned by a required property"""
        self.fspaths = fspaths_converter(self.required_paths())
        return self

    @classmethod
    def checks(cls):
        """Find all methods used to check the validity of the file format

        Returns
        -------
        iter(str)
            an iterator over all method names marked as a "check"
        """
        # Loop through all attributes and find methods marked by CHECK_ANNOTATION
        fileset_props = dir(FileSet)
        for attr_name in dir(cls):
            if attr_name in fileset_props:
                continue
            klass_attr = getattr(cls, attr_name)
            try:
                klass_attr.__annotations__[cls.CHECK_ANNOTATION]
            except (AttributeError, KeyError):
                pass
            else:
                yield attr_name

    def select_by_ext(
        self, fileformat: ty.Optional[type] = None, allow_none: bool = False
    ) -> Path:
        """Selects a single path from a set of file-system paths based on the file
        extension

        Parameters
        ----------
        fileformat : type
            the format class of the path to select
        allow_none : bool
            whether to return None instead of raising an error if the file is not found

        Returns
        -------
        Path or None
            the selected file-system path that matches the extension or None if not found
            and `allow_none` is True

        Raises
        ------
        FileFormatError
            if no paths match the extension and `allow_none` is False
        FileFormatError
            if more than one paths matches the extension
        """
        if fileformat is None:
            fileformat = type(self)
        exts = [fileformat.ext]
        try:
            exts.extend(fileformat.alternate_exts)
        except AttributeError:
            pass
        matches = self.matching_exts(self.fspaths, exts)
        if not matches:
            if allow_none:
                return None
            paths_str = ", ".join(str(p) for p in self.fspaths)
            raise FormatMismatchError(
                f"No matching files with extensions in {exts} in file set {paths_str}"
            )
        elif len(matches) > 1:
            matches_str = ", ".join(str(p) for p in matches)
            raise FormatMismatchError(
                f"Multiple files with {exts} extensions found in : {matches_str}"
            )
        return matches[0]

    @classmethod
    def matching_exts(
        cls, fspaths: ty.Set[Path], exts: ty.Optional[ty.List[str]] = None
    ) -> ty.List[Path]:
        """Returns the paths out of the candidates provided that matches the
        given extension (by default the extension of the class)

        Parameters
        ----------
        fspaths: list[Path]
            The paths to select from
        ext: list[str], optional
            the extensions to match, by default the primary and alternate extensions of
            the class

        Returns
        -------
        list[Path]
            the matching paths

        Raises
        ------
        FileFormatError
            When no paths match or more than one path matches the given extension"""
        if isinstance(fspaths, (str, Path)):
            fspaths = [fspaths]
        if exts is None:
            if cls.ext is None:
                return True
            exts = cls.possible_exts
        return [
            p for p in fspaths if any(e is None or str(p).endswith(e) for e in exts)
        ]

    @classmethod
    def convert(cls, fileset, plugin="serial", task_name=None, **kwargs):
        """Convert a given file-set into the format specified by the class

        Parameters
        ----------
        fileset : FileSet
            the file-set object to convert
        plugin : str
            the "execution plugin" used to run the conversion task
        task_name : str
            the name given to the converter task
        **kwargs
            args to pass to the conversion process

        Returns
        -------
        FileSet
            the file-set converted into the type of the current class
        """
        if task_name is None:
            task_name = f"{type(fileset).__name__}_to_{cls.__name__}_{id(fileset)}"
        # Make unique, yet somewhat recognisable task name
        task = cls.get_converter(source_format=type(fileset), name=task_name, **kwargs)
        result = task(in_file=fileset, plugin=plugin)
        out_file = result.output.out_file
        if not isinstance(out_file, cls):
            out_file = cls(out_file)
        return out_file

    @classmethod
    def get_converter(cls, source_format: type, name: str = "converter", **kwargs):
        """Get a converter that converts from the source format type
        into the format specified by the class

        Parameters
        ----------
        source_format : type
            the format to convert from
        name : str
            the name given to the converter task
        **kwargs
            passed on to the task init method to customise the conversion

        Returns
        -------
        pydra.engine.TaskBase or None
            a pydra task or workflow that performs the conversion, or None if no
            conversion is required

        Raises
        ------
        FileFormatConversionError
            no converters found between source and dest format
        FileFormatConversionError
            ambiguous (i.e. more than one) converters found between source and dest format
        """
        if issubclass(source_format, cls):
            return None
        converters = (
            cls.get_converters_dict()
        )  # triggers loading of standard converters for target format
        try:
            unclassified = source_format.unclassified
        except AttributeError:
            import_extras_module(source_format)
        else:
            import_extras_module(unclassified)
        try:
            converter_tuple = converters[source_format]
        except KeyError:
            # If no direct mapping check for mapping from source super types and wildcard
            # matches
            available_converters = cls.get_converter_tuples(source_format)
            if len(available_converters) > 1:
                available_str = "\n".join(
                    describe_task(a[0]) for a in available_converters
                )
                raise FormatConversionError(
                    f"Ambiguous converters found between '{cls.mime_like}' and "
                    f"'{source_format.mime_like}':\n{available_str}"
                ) from None
            elif not available_converters:
                msg = (
                    f"Could not find converter between '{source_format.mime_like}' and "
                    f"'{cls.mime_like}' formats"
                )
                extras_imported, extras_pkg, extras_pypi = import_extras_module(cls)
                if not extras_imported:
                    msg += (
                        f'. Was not able to import "extras" module, {extras_pkg}, '
                        f"you may want to try installing the '{extras_pypi}' package "
                        f"from PyPI (e.g. pip install {extras_pypi})"
                    )
                raise FormatConversionError(msg) from None
            converter_tuple = available_converters[0]
            # Store mapping for future reference
            converters[source_format] = converter_tuple
        converter, conv_kwargs = converter_tuple
        if kwargs:
            # Merge kwargs provided to get_conveter with stored kwargs
            conv_kwargs = copy(conv_kwargs)
            conv_kwargs.update(kwargs)
        return converter(name=name, **conv_kwargs)

    @classmethod
    def get_converters_dict(cls, klass=None):
        # Only access converters to the specific class, not superclasses (which may not
        # be able to convert to the specific type)
        if klass is None:
            klass = cls
        # import related extras module for the target class
        import_extras_module(klass)
        try:
            converters_dict = klass.__dict__["converters"]
        except KeyError:
            converters_dict = {}
            klass.converters = converters_dict
        return converters_dict

    @classmethod
    def get_converter_tuples(
        cls, source_format: type
    ) -> ty.List[ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]]]:
        """Search the registered converters to find any matches and return list of
        task and associated key-word args to perform the conversion between source and
        target formats

        Parameters
        ----------
        source_format : type(FileSet)
            the source format to convert from

        Returns
        -------
        available : list[tuple[TaskBase, dict[str, Any]]]
            list of available converters between the source and target formats
        """
        converters_dict = cls.get_converters_dict()
        available = []
        for src_frmt, converter in converters_dict.items():
            if len(converter) == 2:  # Ignore converters with wildcards at this point
                if issubclass(source_format, src_frmt):
                    available.append(converter)
        if not available and hasattr(source_format, "unclassified"):
            available = SubtypeVar.get_converter_tuples(
                source_format, target_format=cls
            )
        return available

    @classmethod
    def register_converter(
        cls,
        source_format: type,
        converter_tuple: ty.Tuple[ty.Callable, ty.Dict[str, ty.Any]],
    ):
        """Registers a converter task within a class attribute. Called by the
        @fileformats.mark.converter decorator.

        Parameters
        ----------
        source_format : type
            the source format to register a converter from
        task_spec : ty.Callable
            a callable that resolves to a Pydra task
        converter_kwargs : dict
            additional keyword arguments to be passed to the task spec at initialisation
            time

        Raises
        ------
        FormatConversionError
            if there is already a converter registered between the two types
        """
        converters_dict = cls.get_converters_dict()
        # If no converters are loaded, attempt to load from the standard location
        if source_format in converters_dict:
            prev = cls.converters[source_format][0]
            raise FormatConversionError(
                f"There is already a converter registered between {source_format.__name__} "
                f"and {cls.__name__}: {describe_task(prev)}"
            )
        converters_dict[source_format] = converter_tuple

    @classproperty
    def all_formats(cls) -> set:
        """Iterate over all FileSet formats in fileformats.* namespaces"""
        if cls._all_formats is None:
            cls._all_formats = set(
                f for f in FileSet.subclasses() if f.__dict__.get("iana_mime", True)
            )
        return cls._all_formats

    @classproperty
    def standard_formats(cls):
        """Iterate over all formats in the standard fileformats.* namespaces"""
        return (f for f in cls.all_formats if f.namespace in STANDARD_NAMESPACES)

    @classproperty
    def formats_by_iana_mime(cls):
        """a dictionary containing all formats by their IANA MIME type (if applicable)"""
        if cls._formats_by_iana_mime is None:
            cls._formats_by_iana_mime = {
                f.iana_mime: f
                for f in FileSet.all_formats
                if getattr(f, "iana_mime", None) is not None
            }
        return cls._formats_by_iana_mime

    @classproperty
    def formats_by_name(cls):
        """a dictionary containing lists of formats by their translated class names,
        i.e. their can be more than one format with the same translated name"""
        if cls._formats_by_name is None:
            cls._formats_by_name = {
                k: set(v for _, v in g)
                for k, g in itertools.groupby(
                    sorted(
                        (
                            (to_mime_format_name(f.__name__), f)
                            for f in FileSet.all_formats
                            if getattr(f, "iana_mime", None) is None
                        ),
                        key=itemgetter(0),
                    ),
                    key=itemgetter(0),
                )
            }
        return cls._formats_by_name

    @property
    def all_file_paths(self) -> ty.Iterable[Path]:
        """Paths of all files within the fileset"""
        for fspath in self.fspaths:
            if fspath.is_file():
                yield fspath
            else:
                for dpath, _, file_paths in os.walk(fspath):
                    for file_path in file_paths:
                        yield Path(dpath) / file_path

    def byte_chunks(
        self,
        mtime: bool = False,
        chunk_len=FILE_CHUNK_LEN_DEFAULT,
        relative_to: ty.Optional[os.PathLike] = None,
        ignore_hidden_files: bool = False,
        ignore_hidden_dirs: bool = False,
    ) -> ty.Generator[str, ty.Generator[bytes]]:
        """Yields relative paths for all files within the file-set along with iterators
        over their byte-contents. To be used when generating hashes for the file set.

        Parameters
        ----------
        mtime : bool, optional
            instead of iterating over the entire contents of the file-set, simply yield
            a bytes repr of the last modification time.
        chunk_len : int, optional
            the chunk length to break up the file and calculate the hash over, by default 8192
        relative_to : Path, optional
            the base path by which to record the file system paths in the dictionary keys
            to, by default None
        ignore_hidden_files : bool
            whether to ignore hidden files within nested directories (i.e. those
            starting with '.')
        ignore_hidden_dirs : bool
            whether to ignore hidden directories within nested directories (i.e. those
            starting with '.')

        Yields
        -------
        rel_path: str
            relative path to either 'relative_to' arg or common base path for each file
            in the file set
        byte_iter : Generator[bytes]
            an iterator over the bytes contents of the file, chunked into 'chunk_len'
            chunks
        """
        # If "relative_to" is not provided, get the common path between
        if relative_to is None:
            relative_to = Path(os.path.commonpath(self.fspaths))
            if all(p.is_file() and p.parent == relative_to for p in self.fspaths):
                relative_to /= os.path.commonprefix(
                    [p.name for p in self.fspaths]
                ).rstrip(".")
        # yield the absolute base path if using mtimes instead of contents
        if mtime:
            yield ("<base-path>", iter([str(relative_to.absolute()).encode()]))

        relative_to = str(relative_to)
        if Path(relative_to).is_dir() and not relative_to.endswith(os.path.sep):
            relative_to += os.path.sep

        if mtime:

            def chunk_file(fspath: Path):
                """Yields a byte representation of the last modified time for the file"""
                yield bytes(struct.pack("<d", os.stat(fspath).st_mtime))

        else:

            def chunk_file(fspath: Path):
                """Yields the contents of the file in byte chunks"""
                if not fspath.is_file():
                    assert fspath.is_symlink()  # broken symlink
                    yield b"\x00"
                else:
                    with open(fspath, "rb") as fp:
                        for chunk in iter(functools.partial(fp.read, chunk_len), b""):
                            yield chunk

        def chunk_dir(fspath):
            for dpath, _, filenames in sorted(os.walk(fspath)):
                # Sort in-place to guarantee order.
                filenames.sort()
                dpath = Path(dpath)
                if (
                    ignore_hidden_dirs
                    and dpath.name.startswith(".")
                    and str(dpath) != fspath
                ):
                    continue
                for filename in filenames:
                    if ignore_hidden_files and filename.startswith("."):
                        continue
                    yield (
                        str((dpath / filename).relative_to(relative_to)),
                        chunk_file(dpath / filename),
                    )

        for key, fspath in sorted(
            ((str(p)[len(relative_to) :], p) for p in self.fspaths),
            key=itemgetter(0),
        ):
            if fspath.is_dir():
                yield from chunk_dir(fspath)
            else:
                yield (key, chunk_file(fspath))

    def hash(self, crypto=None, **kwargs) -> bytes:
        """Calculate a unique hash for the file-set based on the relative paths and
        contents of its constituent files

        Parameters
        ----------
        crypto : function, optional
            the cryptography method used to hash the files, by default hashlib.sha256
        **kwargs
            keyword args passed directly through to the ``hash_dir`` function

        Returns
        -------
        hash : str
            unique hash for the file-set
        """
        if crypto is None:
            crypto = hashlib.sha256
        crytpo_obj = crypto()
        for path, bytes_iter in self.byte_chunks(**kwargs):
            crytpo_obj.update(path.encode())
            for bytes_str in bytes_iter:
                crytpo_obj.update(bytes_str)
        return crytpo_obj.hexdigest()

    def hash_files(self, crypto=None, **kwargs) -> dict[str, bytes]:
        """Calculate hashes for all files in the file-set based on the relative paths and
        contents of its constituent files

        Parameters
        ----------
        crypto : function, optional
            the cryptography method used to hash the files, by default hashlib.sha256
        **kwargs
            keyword args passed directly through to the ``hash_dir`` function

        Returns
        -------
        file_hashes : dict[str, bytes]
            unique hashes for each file in the file-set
        """
        if crypto is None:
            crypto = hashlib.sha256
        file_hashes = {}
        for path, bytes_iter in self.byte_chunks(**kwargs):
            crypto_obj = crypto()
            for bytes_str in bytes_iter:
                crypto_obj.update(bytes_str)
            file_hashes[str(path)] = crypto_obj.hexdigest()
        return file_hashes

    def __bytes_repr__(
        self, cache: dict  # pylint: disable=unused-argument
    ) -> ty.Iterable[bytes]:
        """Provided for compatibility with Pydra's hashing function, return the contents
        of all the files in the file-set in chunks

        Parameters
        ----------
        cache : dict
            an object passed around by Pydra's hashing function to store cached versions
            of previously hashed objects, to allow recursive structures

        Yields
        ------
        bytes
            a chunk of bytes of length FILE_CHUNK_LEN_DEFAULT from the contents of all
            files in the file-set.
        """
        cls = type(self)
        yield f"{cls.__module__}.{cls.__name__}:".encode()
        for key, chunk_iter in self.byte_chunks():
            yield (",'" + key + "'=").encode()
            yield from chunk_iter

    @classmethod
    def mock(cls, *fspaths: ty.Tuple[ty.Union[Path, str]]) -> Self:
        """Return an instance of a mocked sub-class of the file format to be used in
        test routines like doctests.

        Parameters
        ----------
        *fspaths: tuple[Path]
            the paths to be provided to the mocked class, by default will be ["mock/<class-name-lower>"]

        Returns
        -------
        FileSet
            a file-set that will pass type-checking as an instance of the given
            fileset class but which doesn't actually point to any FS objects.
        """
        mock_cls = type(cls.__name__ + "Mock", (MockMixin, cls), {"TRUE_CLASS": cls})
        if not fspaths:
            fspaths = []
            fspath = f"/mock/{cls.__name__.lower()}"
            if cls.ext:
                fspath += cls.ext
            fspaths.append(fspath)
        return mock_cls(fspaths=fspaths)

    @classmethod
    def sample(cls, dest_dir: ty.Optional[Path] = None) -> Self:
        """Return an sample instance of the file-set type for classes where the
        `test_data` extra has been implemented

        Parameters
        ----------
        dest_dir : Path, optional
            the path in which to create the test data

        Returns
        -------
        FileSet
            an instance of the given file-set class
        """
        if not dest_dir:
            dest_dir = Path(tempfile.mkdtemp())
        # Need to use mock to get an instance in order to use the singledispatch-based
        # mark.extra decorator
        mock = cls.mock()
        return cls(mock.generate_sample_data(dest_dir))

    @mark.extra
    def generate_sample_data(self, dest_dir: Path) -> ty.Iterable[Path]:
        """Generate test data at the fspaths of the file-set

        Parameters
        ----------
        dest_dir : Path
            the directory to generate the test data within

        Returns
        -------
        fspaths : ty.Iterable
            the generated fspaths
        """
        raise NotImplementedError

    class ExtensionDecomposition(IntEnum):
        """What to consider the file extension to be for paths without an explicitly
        defined extension

        Options
        -------
        none
            assume it doesn't have a file extension, i.e. all parts are included in the stem
        single
            assume that anything after the last '.' is the extension, e.g. the extension
            of "file.nii.gz" would be ".gz"
        multiple
            assume that anything after the first '.' is the extension, e.g. the extension
            of "file.nii.gz" would be "nii.gz"
        """

        none = 0
        single = 1
        multiple = 2

        def __str__(self):
            return self.name

    def decomposed_fspaths(
        self,
        required_only: bool = True,
        decomposition_mode: ExtensionDecomposition = ExtensionDecomposition.single,
    ) -> ty.List[ty.Tuple[Path, str, str]]:
        """Decompose paths into parent directory, filename stem, and extension

        Parameters
        ----------
        required_only : bool, optional
            only include required paths, by default True
        assume_implicit_ext : FileSet.ExtensionDecomposition, optional
            how to interpret paths without an explicitly defined extension (i.e. by
            either the extension of the FileSet or nested filesets), by default single

        Returns
        -------
        decomposed_fspath : list[tuple[Path, str, str]]
            a tuple consisting of the parent directory, file-stem and extension
        """
        from ..generic import File

        decomposed_fspaths = []
        implicit = set(
            self.required_paths()
            if required_only and self.required_paths()
            else self.fspaths
        )
        nested = self.nested_filesets()
        for fileset in [self] + nested:
            if isinstance(fileset, File):
                decomposed = (fileset.fspath.parent, fileset.stem, fileset.actual_ext)
                if fileset.fspath in implicit:
                    decomposed_fspaths.append(decomposed)
                    implicit.remove(fileset.fspath)
                elif decomposed not in decomposed_fspaths:
                    previous_fileset = next(
                        f for f in nested if f.fspath == fileset.fspath
                    )
                    previous = (
                        previous_fileset.fspath.parent,
                        previous_fileset.stem,
                        previous_fileset.actual_ext,
                    )
                    warn(
                        f"The '{fileset.fspath}' path within {self} into has been decomposed as "
                        f"{previous} as it was interpreted as a {type(previous_fileset)} "
                        f"file, whereas it is also interpreted as a {type(fileset)}, in "
                        f"which case it could be alternatively decomposed into {decomposed}"
                    )
        for fspath in implicit:
            decomposed_fspaths.append(
                self.decompose_fspath(fspath, mode=decomposition_mode)
            )
        return decomposed_fspaths

    @classmethod
    def decompose_fspath(
        cls, fspath: Path, mode: ExtensionDecomposition = ExtensionDecomposition.single
    ) -> ty.Tuple[Path, str, str]:
        if isinstance(fspath, str):
            fspath = Path(fspath)
        """Decompose an arbitrary file-system path into parent dir, stem and extension
        given the assumption on what constitutes an extension"""
        if mode == cls.ExtensionDecomposition.multiple:
            ext = "".join(fspath.suffixes)
            stem = fspath.name[: -len(ext)]
        elif mode == cls.ExtensionDecomposition.single:
            stem = fspath.stem
            ext = fspath.suffix
        else:
            assert mode == cls.ExtensionDecomposition.none
            stem = fspath
            ext = ""
        return fspath.parent, stem, ext

    class CopyMode(Enum):
        """Designates the desired behaviour of the FileSet.copy() method with regards to
        symbolic, hard or full copies

        Basic options (in order of preference)
        --------------------------------------
        leave
            simply leave the files where they are (i.e. do nothing)
        symlink
            symlink the files into the destination directory
        hardlink
            hardlink the files into the destination directory
        copy
            duplicate (actually copy) the files into the destination directory

        Common combinations
        -------------------
        link
            use either linking method (preferring symbolic)
        link_or_copy
            use either link method or copy (preferring sym, hard, then copy)
        symlink_or_copy
            "  symbolic "   "         "
        hardlink_or_copy
            "  hard "   "         "

        Masks
        -----
        any
            use any method (preferring leave)
        none
            none of the requested methods are supported (i.e. after masking with the
            supported options mask)
        """

        # Bases

        leave = 0b0001
        symlink = 0b0010
        hardlink = 0b0100
        copy = 0b1000

        # Common combinations

        link = 0b0110
        link_or_copy = 0b1110
        symlink_or_copy = 0b1010
        hardlink_or_copy = 0b1100

        # Masks

        any = 0b1111
        none = 0b0000

        # All other combinations (typically the result of bit-masking)

        leave_or_symlink = 0b0011
        leave_or_hardlink = 0b0101
        leave_or_link = 0b0111
        leave_or_symlink_or_copy = 0b1011
        leave_or_hardlink_or_copy = 0b1101

        def __xor__(self, other):
            return type(self)(self.value ^ other.value)

        def __and__(self, other):
            return type(self)(self.value & other.value)

        def __or__(self, other):
            return type(self)(self.value | other.value)

        def __sub__(self, other):
            return type(self)((self.value & (self.value ^ other.value)))

        def __bool__(self):
            return bool(self.value)

        def __str__(self):
            return self.name

    class CopyCollation(IntEnum):
        """Designates the desired "collation" behaviour of the FileSet.copy() method

        Bases
        -----
        separated
            If mode == leave, paths can exist in separate directories in the
            file-system. For other copy modes, the relative directory structure between
            the "copied" paths (incl. links) in the set will be maintained within the
            destination directory. This guarantees there won't be name clashes between
            copied paths.
        siblings
            copied paths are guaranteed to be "copied" to the root of the destination
            directory, i.e be siblings. However, this requires that the file/dir name
            in fspaths are unique.
        adjacent
            copied paths are guaranteed to have the same name-stem and only differ in
            file-extensions. Requires that the file-set only includes files/dirs with
            unique suffixes (NB: suffixes are considered to be everything after the first
            '.' in the filename).
        """

        any = 0
        siblings = 1
        adjacent = 2

        def __str__(self):
            return self.name

    def copy(
        self,
        dest_dir: Path,
        mode: ty.Union[CopyMode, str] = CopyMode.copy,
        collation: ty.Union[CopyCollation, str] = CopyCollation.any,
        new_stem: ty.Optional[str] = None,
        trim: bool = True,
        make_dirs: bool = False,
        overwrite: bool = False,
        supported_modes: CopyMode = CopyMode.any,
        extension_decomposition: ExtensionDecomposition = ExtensionDecomposition.single,
    ):
        """Copies the file-set to a new directory, optionally renaming the files
        to have consistent name-stems.

        Parameters
        ----------
        dest_dir : str
            Path to the parent directory to save the file-set
        mode : FileSet.CopyMode or str, optional
            designates whether to perform an actual copy or whether a link (symbolic or
            hard) is okay, 'duplicate' by default. See FielSet.CopyMode for details
        collation : FileSet.CopyCollation or str, optional
            how to treat relative paths within the fileset, i.e. whether to move them
            to a single directory, rename them to the same file-stem or maintain
            relative directory structure. See FileSet.CopyCollation for details
        new_stem: str, optional
            the file name excluding file extensions, to give the files/dirs in the parent
            directory, by default the original file name is used
        trim : bool, optional
            Only copy the paths in the file-set that are "required" by the format,
            true by default
        make_dirs : bool, optional
            Make the parent destination and all missing ancestors if they are missing,
            false by default
        overwrite : bool, optional
            whether to overwrite existing files/directories if present
        supported_modes : CopyMode, optional
            supported modes for the copy operation. Used to mask out the requested
            copy mode
        extension_decomposition : FileSet.ExtensionDecomposition, optional
            whether to consider file extensions to start from the first '.' (multiple) or the
            last (single) or be empty (none), when the extension of a fspath in the
            FileSet isn't explicitly defined by the FileSet class. Only relevant when
            collation mode is set to "adjacent". By default True
        """
        mode = self.CopyMode[mode] if isinstance(mode, str) else mode
        selected_mode = mode & supported_modes
        if len(self.fspaths) == 1:
            # If there is only one path to copy, then collation isn't meaningful
            collation = self.CopyCollation.any
        else:
            collation = (
                self.CopyCollation[collation]
                if isinstance(collation, str)
                else collation
            )
        if collation >= self.CopyCollation.siblings:
            if not all(p.parent == self.parent for p in self.fspaths):
                selected_mode -= self.CopyMode.leave
            duplicate_names = [
                n for n, c in Counter(p.name for p in self.fspaths).items() if c > 1
            ]
            if duplicate_names:
                raise FileFormatsError(
                    f"Cannot copy {self} to {dest_dir} with collation mode "
                    f'"{collation}", as there are duplicate filenames, {duplicate_names}, '
                    f"in file paths: " + "\n".join(str(p) for p in self.fspaths)
                )
        if not selected_mode:
            raise FileFormatsError(
                f"Cannot copy {self} using {mode} mode as it is not supported by "
                f"the {supported_modes} given the collation specification, {collation}"
            )
        if selected_mode & self.CopyMode.leave:
            return self  # Don't need to do anything
        if collation == self.CopyCollation.adjacent:
            decomposed_fspaths = self.decomposed_fspaths(
                required_only=trim, decomposition_mode=extension_decomposition
            )
            exts = [d[-1] for d in decomposed_fspaths]
            duplicate_exts = [n for n, c in Counter(exts).items() if c > 1]
            if duplicate_exts:
                raise FileFormatsError(
                    f"Cannot copy {self} to {dest_dir} with collation mode "
                    f'"{collation}", as there are duplicate extensions, {duplicate_exts}, '
                    f"in file paths: " + "\n".join(str(p) for p in self.fspaths)
                )
        dest_dir = Path(dest_dir)  # ensure a Path not a string
        if make_dirs:
            dest_dir.mkdir(parents=True, exist_ok=True)
        if selected_mode & self.CopyMode.symlink:
            copy_dir = copy_file = os.symlink
        elif selected_mode & self.CopyMode.hardlink:
            copy_file = os.link

            def hardlink_dir(src: Path, dest: Path):
                for dpath, _, fpaths in os.walk(src):
                    dpath = Path(dpath)
                    relpath = dpath.relative_to(src)
                    (dest / relpath).mkdir()
                    for fpath in fpaths:
                        os.link(dpath / fpath, dest / relpath / fpath)

            copy_dir = hardlink_dir
        else:
            assert selected_mode & self.CopyMode.copy
            copy_dir = shutil.copytree
            copy_file = shutil.copyfile
        new_paths = []
        if trim and self.required_paths():
            fspaths_to_copy = self.required_paths()
        else:
            fspaths_to_copy = self.fspaths
        if not fspaths_to_copy:
            raise FileFormatsError(
                f"Cannot copy {self} because none of the fspaths in the file-set are "
                "required. Set trim=False to copy all file-paths"
            )
        if collation == self.CopyCollation.adjacent:
            if new_stem is None:
                new_stem = sorted(decomposed_fspaths)[0][1]
            # Warning we redefine fspaths_to_copy as list of tuples not Paths
            fspaths_to_copy = decomposed_fspaths
        elif new_stem is not None:
            raise FileFormatsError(
                f"'stem' ({new_stem}) provided to FileSet.copy() method when collation is "
                f"not set to 'adjacent' ({collation})"
            )
        for fspath in fspaths_to_copy:
            if collation == self.CopyCollation.adjacent:
                # fspath is a path decomposed into parent, stem, ext instead of a Path
                parent_dir, old_stem, ext = fspath
                fspath = parent_dir / (old_stem + ext)  # reconstruct into a Path
                new_path = dest_dir / (new_stem + ext)
            elif collation == self.CopyCollation.siblings:
                new_path = dest_dir / fspath.name
            else:
                assert collation == self.CopyCollation.any
                new_path = dest_dir / fspath.relative_to(self.parent)
                new_path.parent.mkdir(parents=True, exist_ok=True)
            if new_path.exists():
                if overwrite:
                    if fspath.is_dir():
                        shutil.rmtree(new_path)
                    else:
                        os.unlink(new_path)
                else:
                    raise FileFormatsError(
                        f"Destination path '{str(new_path)}' exists, set "
                        "'overwrite' to overwrite it"
                    )
            if fspath.is_dir():
                copy_dir(fspath, new_path)
            else:
                copy_file(fspath, new_path)
            new_paths.append(new_path)
        return type(self)(new_paths)

    def copy_to(self, *args, **kwargs):
        """For b/w compatibility (temporary message)"""
        warn("'FileSet.copy_to()' has been deprecated, please use copy() instead")
        return self.copy(*args, **kwargs)

    _all_formats = None
    _formats_by_iana_mime = None
    _formats_by_name = None


@attrs.define(slots=False, repr=False)
class MockMixin:
    """Strips out validation methods of a class, allowing it to be mocked in a way that
    still satisfies type-checking"""

    # Mirror fspaths here so we can unset its validator
    fspaths: ty.FrozenSet[Path] = attrs.field(default=None, converter=fspaths_converter)

    def __attrs_post_init__(self):
        pass

    @fspaths.validator
    def validate_fspaths(self, _, fspaths):
        pass

    @classproperty
    def _type_name(cls):
        assert cls.__name__.endswith("Mock")
        return cls.__name__[: -len("Mock")]



from fileformats.generic import File


class Foo(File):

    ext = ".foo"
    binary = False


class Bar(File):

    ext = ".bar"
    binary = False


class Baz(File):

    ext = ".baz"
    binary = False


class Qux(File):

    ext = ".qux"
    binary = False



from fileformats.core import from_mime
from fileformats.field import Text, Integer, Decimal, Boolean, Array


def test_field_text_mime():
    assert Text.mime_like == "field/text"
    assert from_mime("field/text") is Text


def test_field_integer_mime():
    assert Integer.mime_like == "field/integer"
    assert from_mime("field/integer") is Integer


def test_field_decimal_mime():
    assert Decimal.mime_like == "field/decimal"
    assert from_mime("field/decimal") is Decimal


def test_field_boolean_mime():
    assert Boolean.mime_like == "field/boolean"
    assert from_mime("field/boolean") is Boolean


def test_field_array_mime():
    assert Array.mime_like == "field/array"
    assert from_mime("field/array") is Array



from fileformats.generic import File
from fileformats.serialization import Json
from fileformats.core.mixin import WithSideCars, WithSeparateHeader


class Y(File):
    ext = ".y"
    binary = False


class Z(File):
    ext = ".z"
    binary = False


class Xyz(WithSideCars, File):

    ext = ".x"
    binary = False
    side_car_types = (Y, Z)


class MyFormat(File):

    ext = ".my"
    binary = False


class MyFormatGz(MyFormat):

    ext = ".my.gz"


class MyFormatX(WithSideCars, MyFormat):

    side_car_types = (Json,)


class YourFormat(File):

    ext = ".yr"
    binary = False


class SeparateHeader(File):

    ext = ".hdr"
    binary = False


class ImageWithHeader(WithSeparateHeader, File):

    ext = ".img"
    header_type = SeparateHeader
    binary = False


class MyFormatGzX(MyFormatX, MyFormatGz):

    pass


class EncodedText(File):
    """A text file where the characters ASCII codes are shifted on conversion
    from text
    """

    ext = ".enc"
    binary = False
